from typing import *
import time

from PyQt5.QtCore import QThread
from base_aux.classes import *


# =====================================================================================================================
pass


# =====================================================================================================================
class ThreadItem(QThread):
    """Object for keeping thread data for better managing.
    """
    target: Callable
    args: Tuple[Any, ...]
    kwargs: Dict[str, Any]

    result: Optional[Any] = None
    exx: Optional[Exception] = None

    def __init__(self, target: Callable, args: Union[Tuple, Any, ] = None, kwargs=None, *_args, **_kwargs):
        super().__init__(*_args, **_kwargs)

        if args and not isinstance(args, (tuple, list, )):
            args = (args, )

        self.target = target
        self.args = args or ()
        self.kwargs = kwargs or {}

    def run(self):
        try:
            self.result = self.target(*self.args, **self.kwargs)
        except Exception as exx:
            msg = f"{exx!r}"
            print(msg)
            self.exx = exx

    def SLOTS_EXAMPLES(self):
        """DON'T START! just for explore!
        """
        # checkers --------------------
        self.started
        self.isRunning()

        self.finished
        self.isFinished()

        self.destroyed
        self.signalsBlocked()

        # settings -------------------
        self.setTerminationEnabled()

        # NESTING --------------------
        self.currentThread()
        self.currentThreadId()
        self.thread()
        self.children()
        self.parent()

        # info --------------------
        self.priority()
        self.loopLevel()
        self.stackSize()
        self.idealThreadCount()

        self.setPriority()
        self.setProperty()
        self.setObjectName()

        self.tr()

        self.dumpObjectInfo()
        self.dumpObjectTree()

        # CONTROL --------------------
        self.run()
        self.start()
        self.startTimer()

        self.sleep(100)
        self.msleep(100)
        self.usleep(100)

        self.wait()

        self.killTimer()

        self.disconnect()
        self.deleteLater()
        self.terminate()
        self.quit()
        self.exit(100)

        # WTF --------------------


# =====================================================================================================================
class ThreadsManager(SingletonByCallMeta):
    """Manager for spawning threads and keep its instances with additional data.
    Singleton! do you dont need saving instances!

    USAGE
    -----
    1. BEST PRACTICE
    Not recommended using it directly, use as simple nested:
        class ThreadsManager1(ThreadsManager):
            pass

        @ThreadsManager1.decorator__to_thread
        def func(*args, **kwargs):
            pass

    2. Direct usage
    But if you need only one manager - do use directly.
        @ThreadsManager.decorator__to_thread
        def func(*args, **kwargs):
            pass

    :ivar _PARAM__NOTHREAD: parameter for passing in decorated function which can run target without thread

    :param args: NAME for manager instance
    :param thread_items: ThreadItem instances,
    :param MUTEX: mutex for safe collecting threads in this manager, creates in init
    :param counter: counter for collected threads in this manager
    """
    THREADS: List[ThreadItem]

    _PARAM__NOTHREAD: str = "nothread"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.THREADS = []

    @property
    def NAME(self) -> str:
        """classname for manager
        """
        return self.__class__.__name__

    @property
    def count(self) -> int:
        return len(self.THREADS)

    # =================================================================================================================
    def decorator__to_thread(self, _func) -> Callable:
        """Decorator which start thread from funcs and methods.

        always collect objects threads in result object! even if nothread! so you can get results from group!

        :param _func: decorated target
        """
        def _wrapper__spawn_thread(*args, **kwargs) -> Optional[Any]:
            """actual wrapper which spawn thread from decorated target.

            :param args: args passed into target/method,
            :param kwargs: kwargs passed into target/method,
            """
            nothread = self._PARAM__NOTHREAD in kwargs and kwargs.pop(self._PARAM__NOTHREAD)

            thread_item = ThreadItem(target=_func, args=args, kwargs=kwargs)
            self.THREADS.append(thread_item)
            thread_item.start()

            if nothread:
                thread_item.wait()
                return thread_item.result

        return _wrapper__spawn_thread

    # =================================================================================================================
    def clear(self) -> None:
        """clear collected thread_items.

        useful if you dont need collected items any more after some step. and need to manage new portion.
        """
        self.THREADS.clear()

    def wait_all(self) -> None:
        """wait while all spawned threads finished.
        """
        # wait all started
        if not self.count:
            time.sleep(0.2)

        for _ in range(3):
            for item in self.THREADS:
                item.wait()

            time.sleep(0.1)

    def terminate_all(self) -> None:
        for thread in self.THREADS:
            thread.terminate()

    def check_results_all(self, value: Any = True, func_validate: Callable[[Any], bool] = None) -> bool:
        """check if result values for all threads are equal to the value

        :param value: expected comparing value for all thread results
        :param func_validate:
        """
        for thread in self.THREADS:
            if func_validate is not None:
                if not func_validate(thread.result):
                    return False
            else:
                if thread.result != value:
                    return False
        return True


# =====================================================================================================================



from typing import *

from . import *


# =====================================================================================================================
# TODO: add Csv???
class PrivateAuto(PrivateJson, PrivateIni, PrivateCsv, PrivateEnv):
    """This class will try all variants in order Json-Ini-Env.
    and take values ONLY from FIRST ONE source with all needed values!
    It will not merge sources!
    """
    def get_dict(self) -> Union[TYPE__PV_DICT, NoReturn]:
        annots = self.annot__get_not_defined()
        annots_lower = set(map(str.lower, annots))

        for cls in [PrivateAuto, PrivateJson, PrivateIni, PrivateCsv, PrivateEnv]:
            try:
                self.FILENAME = super(cls, self).FILENAME
                self._filepath_apply_new()
                data = super(cls, self).get_dict()
                data_lower = set(map(str.lower, data))
                if data_lower.issuperset(annots_lower):
                    return data
            except:
                pass


# =====================================================================================================================
class PrivateAuthAuto(PrivateAuth, PrivateAuto):
    pass


class PrivateTgBotAddressAuto(PrivateTgBotAddress, PrivateAuto):
    pass


# =====================================================================================================================



# DONT RENAME TO SIMPLE "TYPES"!!! it break by circular import


from typing import *

from base_aux.objects.primitives import LAMBDA_TRUE
# LAMBDA_TRUE = lambda *args, **kwargs: True

# =====================================================================================================================
class Cls:
    def meth(self):
        pass


# =====================================================================================================================
TYPE__NONE = type(None)
TYPE__FUNCTION = type(LAMBDA_TRUE)
TYPE__METHOD = type(Cls().meth)


# =====================================================================================================================
@final
class TypeChecker:
    TYPES__ELEMENTARY_SINGLE: tuple = (
        type(None), bool,
        str, bytes,
        int, float,
    )
    TYPES__ELEMENTARY_COLLECTION: tuple = (
        tuple, list,
        set, dict,
    )
    TYPES__ELEMENTARY: tuple = (*TYPES__ELEMENTARY_SINGLE, *TYPES__ELEMENTARY_COLLECTION,)

    # -----------------------------------------------------------------------------------------------------------------
    @staticmethod
    def check__name_is_build_in(name: str) -> bool:
        return name.startswith("__") and name.endswith("__") and len(name) > 4

    # -----------------------------------------------------------------------------------------------------------------
    @staticmethod
    def check__bool_none(source: Any) -> bool:
        """
        GOAL
        ----
        help in case of
            assert 0 == False
            assert 1 == True
            assert 2 == False   # unclear!!!

        CREATED SPECIALLY FOR
        ---------------------
        funcs.Valid.compare_doublesided
        """
        return isinstance(source, (bool, type(None)))

    @staticmethod
    def check__elementary(source: Any) -> bool:
        if callable(source):
            return False
        return isinstance(source, TypeChecker.TYPES__ELEMENTARY)

    @staticmethod
    def check__elementary_single(source: Any) -> bool:
        return isinstance(source, TypeChecker.TYPES__ELEMENTARY_SINGLE)

    @staticmethod
    def check__elementary_single_not_none(source: Any) -> bool:
        """
        its just an Idea!!!

        GOAL
        ----
        prepare to work with ensure_collection
        None assumed as not Passed value! so we can ensure to return None -> ()

        SPECIALLY CREATED FOR
        ---------------------
        ensure_collection somewhere!
        """
        return TypeChecker.check__elementary_single(source) and source is not None

    @staticmethod
    def check__elementary_collection(source: Any) -> bool:
        """
        GOAL
        ----
        MOST PREFERRED to use for ensure_collection! and apply for Args!
        all other objects (ClsInst) will covered by brackets!
        """
        return isinstance(source, TypeChecker.TYPES__ELEMENTARY_COLLECTION)

    @staticmethod
    def check__elementary_collection_not_dict(source: Any) -> bool:
        return isinstance(source, TypeChecker.TYPES__ELEMENTARY_COLLECTION) and not isinstance(source, dict)

    # -----------------------------------------------------------------------------------------------------------------
    @staticmethod
    def check__iterable(
            # self,
            source: Any,
            dict_as_iterable: bool = True,
            str_and_bytes_as_iterable: bool = True,
    ) -> bool:
        """checks if SOURCE is iterable.

        :param source: SOURCE data
        :param dict_as_iterable: if you dont want to use dict in your selecting,
            becouse maybe you need flatten all elements in list/set/tuple into one sequence
            and dict (as extended list) will be irrelevant!
        :param str_and_bytes_as_iterable: usually in data processing you need to work with str-type elements as OneSolid element
            but not iterating through chars!
        """
        if isinstance(source, dict):
            return dict_as_iterable
        elif isinstance(source, (str, bytes)):
            return str_and_bytes_as_iterable
        elif isinstance(source, (tuple, list, set,)):  # need to get it explicitly!!!
            return True
        elif hasattr(source, '__iter__') or hasattr(source, '__getitem__'):
            return True

        # FINAL ---------------------
        return False

    @staticmethod
    def check__iterable_not_str(source: Any) -> bool:
        """
        GOAL
        ----
        checks if SOURCE is iterable, but not exactly str!!!
        """
        return TypeChecker.check__iterable(source, str_and_bytes_as_iterable=False)

    # CALLABLE --------------------------------------------------------------------------------------------------------
    @staticmethod
    def check__callable_func_meth_inst(source: Any) -> bool:
        """
        GOAL
        ----
        just any callable but NO CLASS!!!


        creates specially for detect all funcs like func/meth/or even DescriptedClasses (it is class but actually used like func!)
        recommended using instead of just Callable! cause Callable keeps additionally every class instead of just simple func/method!
        """
        if TypeChecker.check__class(source):
            result = issubclass(source, TypeChecker.TYPES__ELEMENTARY)
        else:
            result = callable(source)
        return result

    @staticmethod
    def check__callable_func_meth(source: Any) -> bool:
        return TypeChecker.check__callable_func(source) or TypeChecker.check__callable_meth(source)

    @staticmethod
    def check__callable_func(source: Any) -> bool:
        """
        if only the exact generic function! no class method!
        Lambda included!
        Classmethod included!

        CREATED SPECIALLY FOR
        ---------------------
        not special! just as ones found ability to!
        """
        if TypeChecker.check__callable_cls_as_func_builtin(source):
            result = True
        else:
            result = type(LAMBDA_TRUE) in source.__class__.__mro__
        return result

    @staticmethod
    def check__callable_meth(source: Any) -> bool:
        """
        if only the exact instance method!
        no generic funcs!
        no CALLABLE INSTANCE!
        no callable classes!

        CREATED SPECIALLY FOR
        ---------------------
        not special! just as ones found ability to!
        """
        result = not TypeChecker.check__class(source) and type(Cls().meth) in source.__class__.__mro__
        return result

    @staticmethod
    def check__callable_inst(source: Any) -> bool:
        """
        CREATED SPECIALLY FOR
        ---------------------
        not special! just as ones found ability to!
        """
        result = TypeChecker.check__instance(source) and hasattr(source, "__call__")
        return result

    @staticmethod
    def check__callable_cls_as_func_builtin(source: Any) -> bool:
        """
        if class and class is as func like int/str/*  or nested
        """
        return TypeChecker.check__class(source) and issubclass(source, TypeChecker.TYPES__ELEMENTARY)

    # CLS/INST --------------------------------------------------------------------------------------------------------
    @staticmethod
    def check__class(source: Any) -> bool:
        """
        works both for funcs/meths for any Сды/Штые1 see tests test__check__class
        """
        # return hasattr(source, "__class__")     # this is incorrect!!! tests get fail!
        try:
            return issubclass(source, object)
        except:
            return False

    @staticmethod
    def check__instance(source: Any) -> bool:
        return not TypeChecker.check__class(source) and not TypeChecker.check__callable_func(source) and not TypeChecker.check__callable_meth(source)

    @staticmethod
    def check__instance_not_elementary(source: Any) -> bool:
        return TypeChecker.check__instance(source) and not TypeChecker.check__elementary(source)

    @staticmethod
    def check__exception(source: Any) -> bool:
        """
        any of both variant (Instance/Class) of any Exception!
        """
        if isinstance(source, Exception):
            return True
        try:
            return issubclass(source, Exception)
        except:
            pass
        return False

    @staticmethod
    def check__nested__by_cls_or_inst(source: Any, parent: Any) -> bool | None:
        """
        any of both variant (Instance/Class) comparing with TARGET of both variant (Instance/Class)

        specially created for pytest_aux for comparing with Exception!
        """
        if TypeChecker.check__instance(source):
            source = source.__class__
        if TypeChecker.check__instance(parent):
            parent = parent.__class__

        if not TypeChecker.check__class(source) or not TypeChecker.check__class(parent):
            return None

        return issubclass(source, parent)


# =====================================================================================================================



from typing import *
import os

from . import TYPE__PV_DICT, PrivateBase


# =====================================================================================================================
class PrivateEnv(PrivateBase):
    """
    read exact environs from Os Environment
    """
    def get_dict(self, _prefix: Optional[str] = None) -> TYPE__PV_DICT:
        """
        directly using result - mainly it is only for PRINTing and debugging! don't use result!

        NOTE: be careful to use result as dict! especially if you have lowercase letters!

        REASON:
            import os

            name_lowercase = "name_lowercase"
            os.environ[name_lowercase] = name_lowercase

            print(os.getenv(name_lowercase))    # name_lowercase
            print(os.getenv(name_lowercase.upper()))    # name_lowercase

            print(os.environ[name_lowercase])   # name_lowercase
            print(os.environ[name_lowercase.upper()])   # name_lowercase

            print(dict(os.environ)[name_lowercase])     # KeyError: 'name_lowercase'
        """
        envs_all = os.environ
        result: TYPE__PV_DICT = {}

        # filter ---------------
        for name, value in envs_all.items():
            if not _prefix or (_prefix and name.upper().startswith(_prefix.upper())):
                result.update({name: value})

        # print ---------------
        print()     # to pretty print in pytest only
        for name, value in result.items():
            print(f"{name}    ={value}")
        print()     # to pretty print in pytest only
        return result


# =====================================================================================================================



from typing import *

from .attr_1_aux import AttrAux
from .static import Exx__AnnotNotDefined


# =====================================================================================================================
class AnnotAux:
    """
    access to all __annotations__
        from all nested classes
        in correct order

    RULES
    -----
    4. nesting available with correct order!
        class ClsFirst(BreederStrStack):
            atr1: int
            atr3: int = None

        class ClsLast(BreederStrStack):
            atr2: int = None
            atr4: int

        for key, value in ClsLast.annotations__get_nested().items():
            print(f"{key}:{value}")

        # atr1:<class 'int'>
        # atr3:<class 'int'>
        # atr2:<class 'int'>
        # atr4:<class 'int'>
    """
    # -----------------------------------------------------------------------------------------------------------------
    def __getattr__(self, item) -> Any | NoReturn:
        return AttrAux._getattr_anycase(item, self)

    def __getitem__(self, item) -> Any | NoReturn:
        return AttrAux._getattr_anycase(item, self)

    # -----------------------------------------------------------------------------------------------------------------
    @classmethod
    def annot__get_nested__dict_types(cls, obj: Any | None = None) -> dict[str, Type[Any]]:
        """
        GOAL
        ----
        get all annotations in correct order (nesting available)!

        RETURN
        ------
        keys - all attr names (defined and not)
        values - Types!!! not instances!!!
        """
        if obj is None:
            obj = cls

        try:
            mro = obj.__mro__
        except:
            mro = obj.__class__.__mro__

        result = {}
        for cls_i in mro:
            if cls_i == AnnotAux or cls_i == object:
                break

            _result_i = dict(cls_i.__annotations__)
            _result_i.update(result)
            result = _result_i
        return result

    def annot__get_nested__dict_values(self, obj: Any | None = None) -> dict[str, Any]:
        """
        GOAL
        ----
        get dict with only existed values! no raise if value not exists!
        """
        if obj is None:
            obj = self

        result = {}
        for key in self.annot__get_nested__dict_types(obj):
            if hasattr(obj, key):
                result.update({key: getattr(obj, key)})
        return result

    def annot__iter_values(self, obj: Any = None):
        yield from self.annot__get_nested__dict_values(obj).values()

    # -----------------------------------------------------------------------------------------------------------------
    def annot__get_not_defined(self, obj: Any | None = None) -> list[str]:
        """
        GOAL
        ----
        return list of not defined annotations

        SPECIALLY CREATED FOR
        ---------------------
        annot__check_all_defined
        """
        if obj is None:
            obj = self

        result = []
        for key in self.annot__get_nested__dict_types(obj):
            if not hasattr(obj, key):
                result.append(key)
        return result

    def annot__check_all_defined(self, obj: Any | None = None) -> bool:
        """
        GOAL
        ----
        check if all annotated attrs have value!
        """
        return not self.annot__get_not_defined(obj)

    def annot__raise_if_not_defined(self, obj: Any | None = None) -> None | NoReturn:
        """
        GOAL
        ----
        check if all annotated attrs have value!
        """
        not_defined = self.annot__get_not_defined(obj)
        if not_defined:
            dict_type = self.annot__get_nested__dict_types(obj)
            msg = f"[CRITICAL]{not_defined=} in {dict_type}"
            raise Exx__AnnotNotDefined(msg)

    # -----------------------------------------------------------------------------------------------------------------
    def annot__print(self, obj: Any = None) -> str:
        """just a pretty print for debugging or research.
        """
        if obj is None:
            obj = self

        result = f"{obj.__class__.__name__}(Annotations):"
        annots = self.annots_get_dict(obj)
        if annots:
            for key, value in annots.items():
                result += f"\n\t{key}={value}"
        else:
            result += f"\nEmpty=Empty"

        return result

    def __str__(self):
        return self.annot__print()


# =====================================================================================================================



from . import *
from typing import *
import re

from . import TYPE__PV_DICT, Exx__SameKeys


# =====================================================================================================================


# =====================================================================================================================
class PrivateCsv(PrivateBase):
    """Get values from CSV file as dict format.

    this is not actually about private values!
    It was created for parsing stdout from CLI commands about device info,
    which is has data like

    skip all nonames.

    skip all lines with no separator.
        C:\\Users\\a.starichenko>STM32_Programmer_CLI --verbosity 1 --connect port=swd index=0
              -------------------------------------------------------------------
                               STM32CubeProgrammer v2.14.0
              -------------------------------------------------------------------

        ST-LINK SN  : 323609013212354D434B4E00
        ST-LINK FW  : V2J29S7
        Board       : --
        Voltage     : 3.26V
        SWD freq    : 4000 KHz
        Connect mode: Normal
        Reset mode  : Software reset
        Device ID   : 0x410
        Revision ID : Rev X
        Device name : STM32F101/F102/F103 Medium-density
        Flash size  : 128 KBytes (default)
        Device type : MCU
        Device CPU  : Cortex-M3
        BL Version  : --


        C:\\Users\\a.starichenko>
    """
    FILENAME = "pv.csv"
    SEPARATOR: str = ":"
    SPACE_IN_KEYS: str = "_"
    RAISE_SAME_KEYS: bool = True
    LINE_SKIP__FIRST: Optional[int] = None
    LINE_SKIP__LAST: Optional[int] = None
    LINE_SKIP__REGEXP: Optional[str] = None

    def get_dict(self) -> Union[TYPE__PV_DICT, NoReturn]:
        result = {}
        lines = self._text.splitlines()[self.LINE_SKIP__FIRST:]
        if self.LINE_SKIP__LAST:
            lines = lines[:-self.LINE_SKIP__LAST]

        for line in lines:
            if self.SEPARATOR not in line:
                continue
            if self.LINE_SKIP__REGEXP and re.search(pattern=self.LINE_SKIP__REGEXP, string=line):
                continue

            key, value = line.split(sep=self.SEPARATOR, maxsplit=1)
            key: str = key.strip()
            key = re.sub(pattern=r"\s", repl=self.SPACE_IN_KEYS, string=key)
            value: str = value.strip()

            if not key:
                continue

            if key in result and self.RAISE_SAME_KEYS:
                raise Exx__SameKeys

            result.update({key: value})

        return result


# =====================================================================================================================
class PrivateAuthCsv(PrivateAuth, PrivateCsv):
    pass


class PrivateTgBotAddressCsv(PrivateTgBotAddress, PrivateCsv):
    pass


# =====================================================================================================================



"""
GOAL
----
collect all obvious variants of code objects

USEFUL IDEAS
------------
- in tests
- could be clearly used in docstrings without needless defining
    assert get_bool(LAMBDA_EXX) is False

"""
# ---------------------------------------------------------------------------------------------------------------------
from typing import *


# =====================================================================================================================
GEN_COMPR = (i for i in range(3))


# ---------------------------------------------------------------------------------------------------------------------
def FUNC(*args, **kwargs) -> None:
    pass


def FUNC_NONE(*args, **kwargs) -> None:
    return None


def FUNC_TRUE(*args, **kwargs) -> bool:
    return True


def FUNC_FALSE(*args, **kwargs) -> bool:
    return False


def FUNC_ALL(*args, **kwargs) -> bool:
    """
    return all(args) and all(kwargs.values())

    CREATED SPECIALLY FOR
    ---------------------
    funcs.Valid.run as tests
    """
    return all(args) and all(kwargs.values())


def FUNC_ANY(*args, **kwargs) -> bool:
    """
    return any(args) or any(kwargs.values())

    CREATED SPECIALLY FOR
    ---------------------
    funcs.Valid.run as tests
    """
    return any(args) or any(kwargs.values())


def FUNC_LIST_DIRECT(*args, **kwargs) -> list[Any]:
    """
    DIRECT LIST() for Args+Kwargs

    CREATED SPECIALLY FOR
    ---------------------
    funcs.Valid.get_bool as test variant

    return list(args) + list(kwargs)
    """
    return list(args) + list(kwargs)


def FUNC_LIST_VALUES(*args, **kwargs) -> list[Any]:
    """
    LIST() values for Args+Kwargs.values()

    CREATED SPECIALLY FOR
    ---------------------
    funcs.Valid.get_bool as test variant

    return list(args) + list(kwargs.values())
    """
    return list(args) + list(kwargs.values())


def FUNC_DICT(*args, **kwargs) -> dict[Any, Any | None]:
    """
    DIRECT DICT() for Args+Kwargs

    CREATED SPECIALLY FOR
    ---------------------
    funcs.Valid.get_bool as test variant

    return like DICT(*args, **kwargs)
    """
    result = dict.fromkeys(args)
    result.update(kwargs)
    return result


def FUNC_EXX(*args, **kwargs) -> NoReturn:
    raise Exception("CALLABLE_EXX")


def FUNC_GEN(*args, **kwargs) -> Generator:
    yield from range(5)


# ---------------------------------------------------------------------------------------------------------------------
LAMBDA = lambda *args, **kwargs: None
LAMBDA_NONE = lambda *args, **kwargs: None
LAMBDA_TRUE = lambda *args, **kwargs: True
LAMBDA_FALSE = lambda *args, **kwargs: False

LAMBDA_ALL = lambda *args, **kwargs: FUNC_ALL(*args, **kwargs)
LAMBDA_ANY = lambda *args, **kwargs: FUNC_ANY(*args, **kwargs)

LAMBDA_LIST_DIRECT = lambda *args, **kwargs: FUNC_LIST_DIRECT(*args, **kwargs)
LAMBDA_LIST_VALUES = lambda *args, **kwargs: FUNC_LIST_VALUES(*args, **kwargs)
LAMBDA_DICT = lambda *args, **kwargs: FUNC_DICT(*args, **kwargs)

# LAMBDA_EXX = lambda *args, **kwargs: raise Exception("LAMBDA_EXX")      # raise=SyntaxError: invalid syntax
LAMBDA_EXX = lambda *args, **kwargs: FUNC_EXX()
# LAMBDA_GEN = lambda *args, **kwargs: yield from range(5)      # yield=SyntaxError: invalid syntax
LAMBDA_GEN = lambda *args, **kwargs: FUNC_GEN()


# =====================================================================================================================
class Exx(Exception):
    pass
INST_EXX = Exx()

# ---------------------------------------------------------------------------------------------------------------------
# class ClsBool(bool):  # cant use it!
#     pass


class ClsInt(int):
    pass


class ClsFloat(float):
    pass


class ClsStr(str):
    pass


class ClsList(list):
    pass


class ClsSet(set):
    pass


class ClsDict(dict):
    pass


# =====================================================================================================================
class Cls:
    pass
INST = Cls()


class ClsEmpty:
    pass
INST_EMPTY = ClsEmpty()


# ---------------------------------------------------------------------------------------------------------------------
class ClsInitArgsKwargs:
    """
    GOAL
    ----
    just apply init with any args/kwargs
    so no Exception would raise in any case!
    in first idea it was not matter to keep them in instance but did it just in case

    CREATED SPECIALLY FOR
    ---------------------
    ClsBoolTrue/*False
    """
    ARGS: tuple
    KWARGS: dict

    def __init__(self, *args, **kwargs):
        self.ARGS = args
        self.KWARGS = kwargs


class ClsInitExx:
    def __init__(self, *args, **kwargs) -> NoReturn:
        raise Exception()


# ---------------------------------------------------------------------------------------------------------------------
class ClsCall:
    def __call__(self, *args, **kwargs) -> None:
        pass

    def meth(self, *args, **kwargs) -> None:
        """
        for other results like None/True/False/Exx use direct LAMBDA/FUNC_*! or wait special necessity.
        """
        pass
INST_CALL = ClsCall()


class ClsCallNone(ClsCall):
    pass
INST_CALL_NONE = ClsCallNone()


class ClsCallTrue:
    def __call__(self, *args, **kwargs) -> bool:
        return True
INST_CALL_TRUE = ClsCallTrue()


class ClsCallFalse:
    def __call__(self, *args, **kwargs) -> bool:
        return False
INST_CALL_FALSE = ClsCallFalse()


class ClsCallExx:
    def __call__(self, *args, **kwargs) -> NoReturn:
        raise Exception()
INST_CALL_EXX = ClsCallExx()


# ---------------------------------------------------------------------------------------------------------------------
class ClsBoolTrue(ClsInitArgsKwargs):
    """
    CREATED SPECIALLY FOR
    ---------------------
    funcs.Valid.get_bool as test variant
    """
    def __bool__(self):
        return True
INST_BOOL_TRUE = ClsBoolTrue()


class ClsBoolFalse(ClsInitArgsKwargs):
    """
    CREATED SPECIALLY FOR
    ---------------------
    funcs.Valid.get_bool as test variant
    """
    def __bool__(self):
        return False
INST_BOOL_FALSE = ClsBoolFalse()


class ClsBoolExx(ClsInitArgsKwargs):
    """
    CREATED SPECIALLY FOR
    ---------------------
    funcs.Valid.get_bool as test variant
    """
    def __bool__(self):
        raise Exception()
INST_BOOL_EXX = ClsBoolExx()


# ---------------------------------------------------------------------------------------------------------------------
class ClsIterYield:
    """
    CONSTRAINTS
    -----------
    YIELD and RETURN all are acceptable!
    several iterations - work fine!

        class Cls:
        def __iter__(self):
            yield from range(3)
            # return iter(range(3))

        obj = Cls()
        for _ in range(2):
            print()
            for i in obj:
                print(i)
    """

    def __iter__(self):
        # RETURN VARIANTS ------------
        # return [1,2,3]  #TypeError: iter() returned non-iterator of type 'list'
        # return range(5)  #TypeError: iter() returned non-iterator of type 'range'
        # return iter([1,2,3])  #OK

        # YIELD VARIANTS ------------    - MOST PREFERRED!(seek would be reset all time to the beginning!!!)
        # yield from [1,2,3]  #OK
        yield from range(5)   #OK
INST_ITER_YIELD = ClsIterYield()


class ClsGen:
    """
    ClsIterNext!
    """
    def __init__(self, start=1, end=3):
        self.start = start
        self.end = end
        self.current = start

    def __iter__(self):
        return self

    def __next__(self):
        if self.current > self.end:
            raise StopIteration
        else:
            result = self.current
            self.current += 1
            return result
INST_GEN = ClsGen()


# ---------------------------------------------------------------------------------------------------------------------
class ClsEq:
    def __init__(self, val: Any = None, *args, **kwargs):
        self.VAL = val

    def __eq__(self, other):
        return other == self.VAL

    def __ne__(self, other):
        return other != self.VAL
INST_EQ = ClsEq()


class ClsEqTrue(ClsInitArgsKwargs):
    def __eq__(self, other):
        raise True
    def __ne__(self, other):
        raise False
INST_EQ_TRUE = ClsEqTrue()


class ClsEqFalse(ClsInitArgsKwargs):
    def __eq__(self, other):
        raise False
    def __ne__(self, other):
        raise True
INST_EQ_FALSE = ClsEqFalse()


class ClsEqExx(ClsInitArgsKwargs):
    def __eq__(self, other):
        raise Exception()

    def __ne__(self, other):
        raise Exception()
INST_EQ_EXX = ClsEqExx()


# ---------------------------------------------------------------------------------------------------------------------
class ClsFullTypes:
    attrZero = 0

    attrNone = None
    attrTrue = True
    attrFalse = False
    attrInt = 1
    attrFloat = 1.1
    attrStr = "str"
    attrBytes = b"bytes"

    attrFunc = FUNC
    attrFuncTrue = FUNC_TRUE
    attrFuncList = FUNC_LIST_DIRECT
    attrFuncDict = FUNC_DICT
    attrFuncExx = FUNC_EXX
    attrFuncGen = FUNC_GEN

    attrGenCompr = GEN_COMPR

    attrCls = ClsEmpty
    attrInst = ClsEmpty()
    attrInstMeth = ClsCall().meth

    attrClsCall = ClsCall
    attrInstCall = ClsCall()
    attrClsCallTrue = ClsCallTrue
    attrInstCallTrue = ClsCallTrue()
    attrClsCallExx = ClsCallExx
    attrInstCallExx = ClsCallExx()

    attrClsIterYield = ClsIterYield
    attrInstIterYield = ClsIterYield()
    attrClsGen = ClsGen
    attrInstGen = ClsGen()

    attrClsBoolTrue = ClsBoolTrue
    attrInstBoolTrue = ClsBoolTrue()
    attrClsBoolFalse = ClsBoolFalse
    attrInstBoolFalse = ClsBoolFalse()
    attrClsBoolExx = ClsBoolExx
    attrInstBooExx = ClsBoolExx()

    attrSet = {1,2,3}
    attrList = [1,2,3]
    attrTuple = (1,2,3)
    attrDict = {1:1}
    attrListInst = [*[Cls(), ] * 3, 1]

    @property
    def propertyNone(self) -> None:
        return
    @classmethod
    @property
    def propertyClassmethodNone(cls) -> None:
        return

    @property
    def propertyInt(self) -> int:
        return 1
    @property
    def propertyExx(self) -> NoReturn:
        raise Exception("exxMsg")
    @property
    def propertyFunc(self) -> Callable:
        return FUNC

    def methNone(self) -> None:
        return
    def methInt(self) -> int:
        return 1
    def methExx(self) -> NoReturn:
        raise Exception("exxMsg")
    def methFunc(self) -> Callable:
        return FUNC
    @classmethod
    def classmethodNone(cls) -> None:
        return
    @staticmethod
    def staticmethodNone() -> None:
        return
INST_FULL_TYPES = ClsFullTypes()


# ---------------------------------------------------------------------------------------------------------------------
CALLABLE_LAMBDA = LAMBDA
CALLABLE_FUNC = FUNC

CALLABLE_CLS = ClsCall
CALLABLE_INST = ClsCall()

CALLABLE_METH_CLS = ClsFullTypes.methNone
CALLABLE_METH_CLS_CLASSMETHOD = ClsFullTypes.classmethodNone
CALLABLE_METH_CLS_STATICMETHOD = ClsFullTypes.staticmethodNone
CALLABLE_METH_CLS_PROPERTY = ClsFullTypes.propertyNone
CALLABLE_METH_CLS_PROPERTY_CLASSMETHOD = ClsFullTypes.propertyClassmethodNone

CALLABLE_METH_INST = ClsFullTypes().methNone
CALLABLE_METH_INST_CLASSMETHOD = ClsFullTypes().classmethodNone
CALLABLE_METH_INST_STATICMETHOD = ClsFullTypes().staticmethodNone
CALLABLE_METH_INST_PROPERTY = ClsFullTypes().propertyNone
CALLABLE_METH_INST_PROPERTY_CLASSMETHOD = ClsFullTypes().propertyClassmethodNone


# =====================================================================================================================



from typing import *
from base_aux.funcs import ValueNotExist


# =====================================================================================================================
class AttrAux:
    """
    NOTICE
    ------
    if there are several same attrs in different cases - you should resolve it by yourself!
    """
    # NAME ------------------------------------------------------------------------------------------------------------
    @classmethod
    def _attr_anycase__find(cls, item: str | Any, obj: Any = ValueNotExist) -> str | None:
        """
        get attr name in original register
        """
        if not isinstance(item, str):
            return

        item = item.strip()
        if obj == ValueNotExist:
            obj = cls   # seems it is not good idea!!!

        for name in dir(obj):
            if name.lower() == str(item).lower():
                return name

        return

    # ATTR ------------------------------------------------------------------------------------------------------------
    @classmethod
    def _getattr_anycase(cls, item: str, obj: Any) -> Any | Callable | NoReturn:
        """
        get attr value by name in any register
        no execution! return pure value as represented in object!
        """
        name_original = cls._attr_anycase__find(item, obj)
        if name_original is None:
            raise AttributeError(item)

        return getattr(obj, name_original)

    @classmethod
    def _setattr_anycase(cls, item: str, value:Any, obj: Any) -> None | NoReturn:
        """
        get attr value by name in any register
        no execution! return pure value as represented in object!

        NoReturn - in case of not accepted names when setattr
        """
        name_original = cls._attr_anycase__find(item, obj)
        if name_original is None:
            if not isinstance(item, str):
                raise AttributeError(item)

            item = item.strip()
            if item in ["", ]:
                raise AttributeError(item)
            name_original = item

        # NOTE: you still have no exx with setattr(obj, "    HELLO", value) and ""
        setattr(obj, name_original, value)

    # ITEM ------------------------------------------------------------------------------------------------------------
    @classmethod
    def _getitem_anycase(cls, item: str, obj: Any) -> Any | Callable | NoReturn:
        return cls._getattr_anycase(item, obj)

    @classmethod
    def _setitem_anycase(cls, item: str, value: Any, obj: Any) -> None | NoReturn:
        cls._setattr_anycase(item, value, obj)


# =====================================================================================================================



# =========================================================================================
### 0. BEST PRACTICE
from base_aux.alerts import AlertSelect


class AlertADX(AlertSelect.TELEGRAM_DEF):
    pass

AlertADX("hello")
AlertADX("World")
AlertADX.threads_wait_all()

# =========================================================================================
# =========================================================================================
# =========================================================================================
### AlertSmtp
#### 1. add new server if not exists
from base_aux.alerts import *


class SmtpServersMOD(SmtpServers):
    EXAMPLE_RU: SmtpAddress = SmtpAddress("smtp.EXAMPLE.ru", 123)


class AlertSmtpMOD(AlertSmtp):
    SERVER_SMTP: SmtpAddress = SmtpServersMOD.EXAMPLE_RU  # or direct =SmtpAddress("smtp.EXAMPLE.ru", 123)

# =========================================================================================
#### 2. change authorisation data (see `privates` for details)
from base_aux.alerts import *
from base_aux.privates import *


class AlertSmtpMOD(AlertSmtp):
    AUTH: PrivateAuto = PrivateAuto(_section="AUTH_EMAIL_MOD")

# =========================================================================================
#### 3. change other settings (see source for other not mentioned)
from base_aux.alerts import *


class AlertSmtpMOD(AlertSmtp):
    RECONNECT_PAUSE: int = 60
    RECONNECT_LIMIT: int = 10

    TIMEOUT_RATELIMIT: int = 600

    RECIPIENT_SPECIAL: str = "my_address_2@mail.ru"

# =========================================================================================
#### 4. send
# if no mods
from base_aux.alerts import *

AlertSmtp(_subj_name="Hello", body="World!")

# with mods
from base_aux.alerts import *


class AlertSmtpMOD(AlertSmtp):
    pass  # changed


AlertSmtpMOD(_subj_name="Hello", body="World!")

# =========================================================================================
#### 5. using in class with saving alert object
from base_aux.alerts import *

class AlertSmtpMOD(AlertSmtp):
    pass    # changed

class MyMonitor:
    ALERT = AlertSmtpMOD

monitor = MyMonitor()
monitor.ALERT("Hello")

# =========================================================================================
# =========================================================================================
### AlertTelegram
# All idea is similar to AlertSmtp.

# add auth data
# add pv.json or do smth else (for details see privates.PrivateJsonTgBotAddress)
# json
{
    "TG_ID": {"MyTgID": 1234567890},
    "TGBOT_DEF": {
        "LINK_ID": "@my_bot_20230916",
        "NAME": "my_bot",
        "TOKEN": "9876543210xxxxxxxxxxxxxxxxxxxxxxxxx"
    }
}

# =========================================================================================
from base_aux.alerts import *

class MyMonitor:
    ALERT = AlertTelegram

monitor = MyMonitor()
monitor.ALERT("Hello")



from typing import *
import json


# =====================================================================================================================
TYPES_ELEMENTARY_SINGLE: tuple = (
    type(None), bool,
    str, bytes,
    int, float,
)
TYPES_ELEMENTARY_COLLECTION: tuple = (
    tuple, list,
    set, dict,
)
TYPES_ELEMENTARY: tuple = (*TYPES_ELEMENTARY_SINGLE, *TYPES_ELEMENTARY_COLLECTION,)

TYPE_ELEMENTARY = Union[*TYPES_ELEMENTARY]


# =====================================================================================================================
class Strings:
    SOURCE: str

    def __init__(self, source: Optional[str] = None):
        self.SOURCE = source

    def try_convert_to__elementary(self, source: Optional[Any] = None) -> TYPE_ELEMENTARY:
        """
        by now it works correct only with single elementary values like INT/FLOAT/BOOL/NONE
        for collections it may work but may not work correctly!!! so use it by your own risk and conscious choice!!
        """
        # FIXME: this is not work FULL and CORRECT!!!! need FIX!!!

        # INIT source -------------
        if source is None:
            source = self.SOURCE

        # PREPARE SOURCE ----------
        source_original = source
        if isinstance(source, str):
            # convert to jason expected - VALUES FOR NULL/FALSE/TRUE
            source = source.replace("True", "true")
            source = source.replace("False", "false")
            source = source.replace("None", "null")

        # WORK --------------------
        try:
            source_elementary = json.loads(source)
            return source_elementary
        except Exception as exx:
            print(f"{exx!r}")
            return source_original


# =====================================================================================================================



from typing import *


# =====================================================================================================================
TYPE__IO = tuple[str, list[str]]


# =====================================================================================================================
# TODO:
#   USE OBJECTS instead of tuples? + internal add+read
#   1. use deque with objects Write Read WEol REol None(not used)?
#   2. add top(n=10)
#   if when writeLine - always readLines?


class HistoryIO:
    history: list[TYPE__IO] = []

    def __init__(self):
        self.history = []

    def count(self) -> int:
        """
        count requests
        """
        return len(self.history)

    def clear(self) -> None:
        self.history.clear()

    # ADD =============================================================================================================
    def add_input(self, data: str) -> None:
        self.history.append((data, []))

    def add_output(self, data: Union[str, list[str]]) -> None:
        if not self.history:
            self.add_input("")

        if isinstance(data, (tuple, list, )):
            self.last_outputs.extend(data)
        else:
            # SINGLE
            self.last_outputs.append(data)

    def add_io(self, data_i: str, data_o: Union[str, list[str]]) -> None:
        self.add_input(data_i)
        self.add_output(data_o)

    def add_history(self, history: 'HistoryIO') -> None:
        for data_i, data_o in history.history:
            self.add_input(data_i)
            self.add_output(data_o)

    # LAST ============================================================================================================
    @property
    def last_io(self) -> TYPE__IO | None:
        try:
            return self.history[-1]
        except:
            return None

    @property
    def last_input(self) -> str:
        try:
            return self.last_io[0]
        except:
            return ""

    @property
    def last_outputs(self) -> list[str]:
        try:
            return self.last_io[1]
        except:
            return []

    @property
    def last_output(self) -> str:
        try:
            return self.last_outputs[-1]
        except:
            return ""

    # CHECK ===========================================================================================================
    def check_equal_io(self) -> bool:
        """
        CREATED SPECIALLY FOR
        ---------------------
        testing UART bus when shorted Rx+Tx
        """
        for data_i, data_o in self.history:
            if not (len(data_o) == 1 and data_o[0] == data_i):
                return False
        return True

    # LIST+PRINT ======================================================================================================
    def list_input(self) -> list[str]:
        result = []
        for data_i, data_o in self.history:
            result.append(data_i)
            print(data_i)
        return result

    def list_output(self) -> list[str]:
        result = []
        for data_i, data_o in self.history:
            for line in data_o:
                result.append(line)
                print(line)
        return result

    def print_io(self) -> None:
        name = "print_io"
        print("="*10 + f"{name.upper():=<90}")
        print()
        for data_i, data_o in self.history:
            print(f"{data_i:20}:", end="")
            indent = ""
            if data_o:
                for line in data_o:
                    line = f"{indent}{line}"
                    print(line)
                    indent = " "*20 + ":"
            else:
                print()
        print("="*100)

    def as_dict(self) -> dict[str, list[str]]:
        """not correct if exists same write"""
        return dict(self.history)


# =====================================================================================================================



from typing import *
from base_aux.objects import TypeChecker
from . import TYPE__VALID_ARGS, Default, ArgsEmpty


# =====================================================================================================================
def args__ensure_tuple(args: TYPE__VALID_ARGS = (), none_as_empty: bool | None = None) -> tuple:
    """
    GOAL
    ----
    used in methods which handle with *args/args (like in funcs.Valid)
    when we want to apply singular value without tuple

    CREATED SPECIALLY FOR
    ---------------------
    Valid
    but can be used in any funcs!

    NOTE
    ----
    dont disclose iterables!!!

    USAGE
    -----
        def func1(*args):
            args = ensure_tuple(args)

        def func2(link: Callable, args):
            args = ensure_tuple(args)
            result = link(*args)

    :param args:
        NONE - is equivalent to SingleElement! so None -> (None, )
        any elementaryCollection - will convert to tuple!
        for others (classes/instances/irerables/generators/...) will assumed as single Object!!! and applied in tuple!!!

        unpacked iterables/generators - if need unpack it manually!!! its not difficult and so clear!
        elementary collection would unpack!

    :param none_as_empty:
        use carefully!

    """
    # TODO: move to object-info or funcsAux???

    # None --------------------------
    if args is None and none_as_empty:
        return ()

    # APPLY DEFAULT --------------------------
    if args is Default or args is ArgsEmpty:    # use only IS!
        args = ()
    elif isinstance(args, Default):
        args = args()

    # ENSURE TUPLE --------------------------
    if not TypeChecker.check__elementary_collection(args):
        result = (args, )
    else:
        result = tuple(args)
    return result


# =====================================================================================================================
def ensure_class(source: Any) -> type:
    """
    GOAL
    ----
    get class from any object

    CREATED SPECIALLY FOR
    ---------------------
    classes.ClsMiddleGroup
    """
    if TypeChecker.check__class(source):
        return source
    else:
        return source.__class__


# =====================================================================================================================



from typing import *
from base_aux.funcs.value_0_explicit import Explicit


# =====================================================================================================================
TYPE__ITERABLE_PATH_KEY = Union[str, int]
TYPE__ITERABLE_PATH_ORIGINAL = List[TYPE__ITERABLE_PATH_KEY]
TYPE__ITERABLE_PATH_EXPECTED = Union[TYPE__ITERABLE_PATH_KEY, TYPE__ITERABLE_PATH_ORIGINAL]
TYPE__ITERABLE = Union[dict, list, tuple, set, Iterable]


# =====================================================================================================================
class IterAux:
    """
    collect universal funcs which work with collections
    """
    # AUX ---------------------
    DATA: TYPE__ITERABLE
    PATH: List[TYPE__ITERABLE_PATH_KEY]

    def __init__(self, data: Optional[TYPE__ITERABLE] = None):
        self.DATA = data or {}
        self.PATH = []

    @classmethod
    def item__get_original__case_insensitive(
            cls,
            item_expected: Any,
            source: TYPE__ITERABLE
    ) -> Optional[Explicit]:
        """
        get FIRST original item from any collection by comparing str(expected).lower()==str(original).lower().

        NOTE:
        1. NONE RESULT__VALUE - RESOLVED!!!
        2. SEVERAL VALUES - not used! by now it is just FIRST matched!
            several items? - it is not useful!!! returning first is most expected!

        USEFUL in case-insensitive systems (like terminals or serial devices) or object structured by prefix-names:
        1. get key in dict
        2. find attribute name in objects

        :param source:
        :param item_expected:
        :return: actual item from collection
            None - if VALUE is unreachable
        """
        for value in list(source):
            if str(value).lower() == str(item_expected).lower():
                return Explicit(value)

    def path__get_original(
            self,
            path_expected: TYPE__ITERABLE_PATH_EXPECTED,
            data: Optional[TYPE__ITERABLE] = None,
    ) -> Optional[Explicit]:
        """
        NOTES:
        1. path used as address KEY for dicts and as INDEX for other listed data
        2. separator is only simple SLASH '/'!

        :param data:
        :param path_expected:
        :return:
            None - if path is unreachable/incorrect
            List[Any] - reachable path which could be used to get VALUE from data by chain data[i1][i2][i3]
        """
        if data is None:
            data = self.DATA

        # prepare type ----------------------------
        if isinstance(path_expected, (list, tuple, set)):
            path_expected = list(path_expected)
        else:
            path_expected = str(path_expected)

        if isinstance(path_expected, str):
            path_expected = path_expected.split("/")

        # work ----------------------------
        path_original = []
        for path_part in path_expected:
            if isinstance(data, dict):
                # DICT ----------------
                address_original = self.item__get_original__case_insensitive(path_part, data)
                if not address_original:
                    return
                address_original = address_original()
                data = data[address_original]

            elif isinstance(data, (list, tuple)):
                # ITERABLE ----------------
                try:
                    address_original = int(path_part)
                    data = data[address_original]
                except:
                    return
            else:
                return
            path_original.append(address_original)

        return Explicit(path_original)

    def value_by_path__get(
            self,
            path_expected: TYPE__ITERABLE_PATH_EXPECTED,
            data: Optional[TYPE__ITERABLE] = None
    ) -> Optional[Explicit]:
        if data is None:
            data = self.DATA

        # work ----------------------------
        path_original = self.path__get_original(path_expected, data)
        try:
            for path_part in path_original():
                data = data[path_part]
        except:
            return

        return Explicit(data)

    def value_by_path__set(
            self,
            path_expected: TYPE__ITERABLE_PATH_EXPECTED,
            value: Any,
            data: Optional[TYPE__ITERABLE] = None
    ) -> bool:
        if data is None:
            data = self.DATA

        # work ----------------------------
        path_original = self.path__get_original(path_expected, data)
        try:
            length = len(path_original())
            for pos, path_part in enumerate(path_original(), start=1):
                if pos == length:
                    data[path_part] = value
                else:
                    data = data[path_part]
        except:
            return False

        return True


# =====================================================================================================================



import time
from typing import *

from base_aux.classes import Lambda, AttrLambdaCall
from base_aux.privates import *
from .base import *


# =====================================================================================================================
class RecipientTgID(PrivateAuto):
    """Object to get telegram RecipientId
    """
    SECTION = "TG_ID"
    MyTgID: str


# =====================================================================================================================
class AlertTelegram(AttrLambdaCall, AlertBase):
    """realisation for sending Telegram msg
    """
    # SETTINGS ------------------------------------
    SERVER_TG: PrivateTgBotAddressAuto = Lambda(PrivateTgBotAddressAuto, _section="TGBOT_DEF")

    # AUX -----------------------------------------
    _conn: telebot.TeleBot

    def _connect_unsafe(self) -> Union[bool, NoReturn]:
        self._conn = telebot.TeleBot(token=self.SERVER_TG.TOKEN)
        return True

    def _send_unsafe(self) -> Union[bool, NoReturn]:
        self._conn.send_message(chat_id=self.RECIPIENT, text=self._msg_compose())
        return True

    def _msg_compose(self) -> str:
        msg = f"{self.SUBJECT}\n{self.body}"
        return msg

    def _recipient_self_get(self) -> str:
        return RecipientTgID().MyTgID


# =====================================================================================================================



from base_aux.threads import *
import time


# =====================================================================================================================
count = 5
time_start = time.time()


# define victim ------------------
class ThreadManager1(ThreadsManager):
    pass


class Cls:
    @ThreadManager1().decorator__to_thread
    def func1(self, num):
        time.sleep(1)
        return num * 1000


# spawn ------------------
for i in range(count):
    assert Cls().func1(i) is None

assert ThreadManager1().count == count
ThreadManager1().wait_all()
assert {item.result for item in ThreadManager1().THREADS} == {num * 1000 for num in range(count)}

ThreadManager1().clear()

# spawn ------------------
for i in range(count):
    assert Cls().func1(i) is None


# =====================================================================================================================



# =====================================================================================================================
# DONOT apply PrivateBase nesting!

class PrivateAuth:
    """Typical structure for AUTH

    :ivar USER: user login name
    :ivar PWD: password
    """
    USER: str
    PWD: str


class PrivateTgBotAddress:
    """Typical structure for Telegram bot address

    :ivar LINK_ID: just a bot id, not important
    :ivar NAME: just a bot public name, not important
    :ivar TOKEN: bot token for connection, important!
    """
    LINK_ID: str     # @mybot20230913
    NAME: str        # MyBotPublicName
    TOKEN: str


# =====================================================================================================================



from typing import *
from pydantic import BaseModel


# =====================================================================================================================
TYPES__DICT = dict[str, Union[None, str, bool, int, float, dict, list]]


# =====================================================================================================================
class ModelStandInfo(BaseModel):
    STAND_NAME: str           # "StandPSU"
    STAND_DESCRIPTION: str    # "test PSU for QCD"
    STAND_SN: str
    STAND_SETTINGS: TYPES__DICT = {}     # main settings for all TCS


class ModelDeviceInfo(BaseModel):
    DUT_INDEX: int          # device position in stand

    DUT_NAME: str           # "PSU"
    DUT_DESCRIPTION: str    # "Power Supply Unit"
    DUT_SN: str


class ModelTcInfo(BaseModel):
    TC_NAME: str
    TC_DESCRIPTION: str

    TC_ASYNC: bool
    TC_SKIP: bool

    TC_SETTINGS: TYPES__DICT = {
        # CONTENT IS NOT SPECIFIED!
        # "ANY_1": Any,
    }


class ModelTcResult(BaseModel):
    tc_timestamp: float | None = None

    tc_active: bool = False
    tc_progress: int = 0
    tc_result: bool | None = None
    tc_details: TYPES__DICT = {
        # CONTENT IS NOT SPECIFIED!
        # "ANY_2": Any,
    }


# =====================================================================================================================
class ModelTcResultFull(ModelTcResult, ModelTcInfo, ModelDeviceInfo):
    pass


class ModelTpInfo(ModelStandInfo):
    TESTCASES: list[ModelTcInfo]


class ModelTpResults(ModelStandInfo):
    TESTCASES: list[list[ModelTcResultFull]]


# =====================================================================================================================



# NOTICE:
# 1. If bus cmd return several lines (DUMP for example) - you will get all of them in list! 
# 2. All answers you will get as string and you must parse it by youself!  
# errors will be get within it.

from typing import *
from base_aux.buses import *

# SHOW (optional) COMMANDS EXPLICITLY by annotations without values!
# ------------------------------------------------------------------
class MySerialDevice(SerialClient):
    IDN: Callable[[Any], TYPE__RW_ANSWER]
    ADDR: Callable[[Any], TYPE__RW_ANSWER]
    DUMP: Callable[[Any], TYPE__RW_ANSWER]

# USE in code
# -----------
dev = MySerialDevice()
if dev.connect():
    answer1 = dev.IDN()   # return answer for sent string in port "IDN"
    answer2 = dev.VIN()   # return answer for sent string in port "VIN"
    answer3 = dev.VIN(12)   # return answer for sent string in port "VIN 12"
    answer4 = dev.VIN("12")   # return answer for sent string in port "VIN 12"




from base_aux.buses.serial_client import SerialClient
SerialClient.addresses_paired__detect()



exit()
from base_aux.buses import *
from typing import *


class Dev(SerialClient_FirstFree_Shorted):
    BAUDRATE: int = 115200
    EOL__SEND = b"\r\n"
    RAISE_CONNECT = False
    LOG_ENABLE = True

    REWRITEIF_READFAILDECODE = 5
    # TIMEOUT__READ: float = 3       # 0.2 is too short!!! dont touch! in case of reading char by char 0.5 is the best!!! 0.3 is not enough!!!

    # def address__validate(self) -> bool | None | NoReturn:
    #     for i in range(5):
    #         self._write('help')
    #         if self.read_line():
    #             break
    #         time.sleep(1)
    #     self.buffers_clear()


dev = Dev()
if dev.connect():
    # for i in range(10):
    #     if dev.write_read__last('stop'):
    #         break
    #     time.sleep(0.5)
    #     print()
    #     print()
    #     print()
    #     print()
    #     print()


    index = 0
    while True:
        index += 1
        load = f"step{index}"
        print(load)
        if not dev.write_read__last_validate(load, load):
            print(f"FAIL_FINAL[{load=}]")



from . import *
from typing import *
from configparser import ConfigParser

from . import TYPE__PV_DICT


# =====================================================================================================================
class PrivateIni(PrivateBase):
    """Get values from Ini file.
    Not recommended using DEFAULT SECTION!
    """
    FILENAME: str = "pv.ini"

    def get_dict(self) -> Union[TYPE__PV_DICT, NoReturn]:
        ini = ConfigParser()

        try:
            ini.read_string(self._text)
        except Exception as exx:
            msg = f"[CRITICAL] incorrect format file!\n{exx!r}"
            print(msg)
            raise exx

        if not self.SECTION or self.SECTION == "DEFAULT" or ini.has_section(section=self.SECTION):
            result = dict(ini[self.SECTION or "DEFAULT"])
            return result
        else:
            msg = f"[CRITICAL] NO [{self.SECTION=} in {self.filepath=}]\n"
            msg += self._text
            print(msg)


# =====================================================================================================================
class PrivateAuthIni(PrivateAuth, PrivateIni):
    pass


class PrivateTgBotAddressIni(PrivateTgBotAddress, PrivateIni):
    pass


# =====================================================================================================================



import time
from typing import *
import threading
import re

import imaplib
import email

from base_aux.classes import Lambda, AttrLambdaCall
from base_aux.privates import *
from base_aux.alerts import *


# =====================================================================================================================
# TODO: SEPARATE BASE+INTERFACE!
# TODO: realise MARK_AS_READ = False!
# TODO: start reading from last read not just unseen!


# =====================================================================================================================
class ImapAddress(NamedTuple):
    """class for keeping connection parameters/settings for exact imap server

    :ivar ADDR: server address like "imap.mail.ru"
    :ivar PORT: server port like 993
    """
    ADDR: str
    PORT: int


class ImapServers:
    """well known servers addresses.

    Here we must collect servers like MilRu/GmailCom, and not to create it in any new project.
    """
    MAIL_RU: ImapAddress = ImapAddress("imap.mail.ru", 993)


class AlertImap(AlertSelect.TELEGRAM_DEF):
    """Just created exact class for Alerts in Imap!
    """
    pass


# =====================================================================================================================
class MonitorImap(AttrLambdaCall, threading.Thread):
    """Monitor (threaded) email box for new letters. Notify if new letter appears corresponding to folder name and subject regexp.

    :ivar INTERVAL: monitoring interval in seconds
    :ivar SERVER: server address
    :ivar AUTH: server authentication data object
    :ivar FOLDER: email folder name where wait new emails
        None - for Inbox/Входящие!
    :ivar SUBJECT_REGEXP: regexp for email subject
        None - for all,
        example - r"\[ALERT\]test1"
    # :ivar MARK_AS_READ: after reading mark as read - NOT REALISED!
    :ivar ALERT: object which will send the alerts
    :ivar _conn: connection object
    :ivar stop_flag: flag for stop monitoring, now is stops only after calculating
    :ivar step_counter: just a counter for actually finished CALCULATING cycles

    :ivar _detected: detected email objects
    """
    INTERVAL: int = 1 * 1 * 10

    SERVER: ImapAddress = ImapServers.MAIL_RU
    AUTH: PrivateBase = Lambda(PrivateAuthAuto, _section="AUTH_EMAIL_DEF")
    FOLDER: Optional[str] = None
    SUBJECT_REGEXP: Optional[str] = None
    # MARK_AS_READ: bool = True

    ALERT: Type[AlertBase] = AlertImap
    _conn: Optional[imaplib.IMAP4_SSL] = None
    stop_flag: Optional[bool] = None
    step_counter: int = 0

    def __init__(self):
        super().__init__(daemon=True)

        self._detected: List[str] = []
        self.start()

    # CONNECT =========================================================================================================
    def _connect(self) -> Union[bool, NoReturn]:
        """connect, create connection object and establish connection.
        """
        if self._conn:
            return True

        print(f"\n _connect {self.__class__.__name__}")
        try:
            self._conn = imaplib.IMAP4_SSL(self.SERVER.ADDR, self.SERVER.PORT)
        except:
            pass

        if self._conn:
            print(self._conn.login(self.AUTH.USER, self.AUTH.PWD))
            self.folder_select(self.FOLDER)
            print(self._conn.search(None, 'UNSEEN'))  # ('OK', [b''])

        return bool(self._conn)

    def folder_select(self, name: Optional[str] = None) -> Optional[NoReturn]:
        """Select folder for monitoring
        """
        try:
            if name:
                print(self._conn.select(name))  # ('OK', [b'5'])
            else:
                print(self._conn.select())  # ('OK', [b'5'])

        except Exception as exx:
            msg = f"[CRITICAL] not exists [folder={name}]"
            print(msg)
            raise Exception(msg)
            # raise exx

    def _conn_close(self) -> None:
        """close connection
        """
        if self._conn:
            self._conn.close()
            self._conn.logout()

    def _conn_clear(self) -> None:
        """delete conn object
        """
        self._conn = None

    def _conn_check_empty(self) -> bool:
        """check if connection obj exists
        """
        return self._conn is None

    # START ===========================================================================================================
    def run(self):
        """MAIN function working in thread
        """
        while True:
            if self._connect():
                subjects = self.get_unseen_subject_list()
                for subject in subjects:
                    if not self.SUBJECT_REGEXP or re.fullmatch(self.SUBJECT_REGEXP, subject):
                        self._detected.append(subject)
                        self.ALERT(subject, subj_suffix=f"{self.__class__.__name__}/{self.FOLDER or 'Inbox'}")

            self._conn_close()
            self.step_counter += 1

            if self.stop_flag:
                return
            time.sleep(self.INTERVAL)
            if self.stop_flag:
                return

    def wait_cycle(self, sleep: int = 1) -> None:
        """wait finished calculated cycles

        :param sleep: sleep period in seconds between checks
        """
        step_finish = self.step_counter + 1

        while self.step_counter < step_finish:
            time.sleep(sleep)

    # MAIL ============================================================================================================
    def get_unseen_subject_list(self) -> List[str]:
        """get all subgects for unseen emails
        """
        result = []

        try:
            for num in self._conn.search(None, "UNSEEN")[1][0].split():    # UNSEEN/ALL
                _, msg = self._conn.fetch(num, '(RFC822)')
                msg = email.message_from_bytes(msg[0][1])

                try:
                    # has russian
                    subject = email.header.decode_header(msg["Subject"])[0][0].decode()
                except Exception as exx:
                    # english only!
                    subject = msg["Subject"]

                subject = subject or ""
                result.append(subject)
        except Exception as exx:
            self._conn_clear()

        return result

    # UNSORTED ========================================================================================================
    def _data_blocks__print(self):
        """just an explorer and research
        """
        _, msg = self._conn.fetch("6", '(RFC822)')
        msg = email.message_from_bytes(msg[0][1])
        print(type(msg))  # <class 'email.message.Message'>
        print(list(
            msg))  # ['Delivered-To', 'Return-path', 'Received-SPF', 'Received', 'Received', 'DKIM-Signature', 'DKIM-Signature', 'Precedence', 'List-Id', 'List-Issue', 'List-Unsubscribe', 'List-Subscribe', 'List-Archive', 'List-Post', 'X-Felis-L', 'X-rpcampaign', 'X-Mailru-Msgtype', 'Feedback-Id', 'Message-Id', 'Date', 'From', 'To', 'Subject', 'MIME-Version', 'Content-Type', 'Content-Transfer-Encoding', 'X-Mailru-Src', 'X-4EC0790', 'X-6b629377', 'X-7564579A', 'X-77F55803', 'X-7FA49CB5', 'X-C1DE0DAB', 'X-C8649E89', 'X-D57D3AED', 'X-F696D7D5', 'X-Mailru-BIMI-Organization', 'X-Mailru-Dmarc-Auth', 'X-Mailru-ThreadID', 'X-Mras', 'X-Spam', 'Authentication-Results', 'X-Mailru-Intl-Transport']

        for name in list(msg):
            print(f"[{name}]===[{msg[name]}]")


# =====================================================================================================================



from base_aux.requirements import *


ReqCheckStr_Os().bool_if__WINDOWS()
ReqCheckStr_Os().bool_if_not__WINDOWS()
ReqCheckStr_Os().raise_if__LINUX()


class ReqCheckStr_Os_MY(ReqCheckStr_Os):
    LINUX: bool = True
    WINDOWS: bool = False


ReqCheckStr_Os_MY()  # check requirement!

Packages().upgrade_pip()



from typing import *

from base_aux.objects import TypeChecker
from base_aux.funcs import *


# =====================================================================================================================
class ValidAux:
    """
    Try to keep all validating funcs in separated place
    """
    # -----------------------------------------------------------------------------------------------------------------
    @classmethod
    def get_result_or_raise(
            cls,
            source: TYPE__VALID_SOURCE,
            args: TYPE__VALID_ARGS = None,
            kwargs: TYPE__VALID_KWARGS = None,
    ) -> Any | NoReturn:
        """
        SPECIFIC LOGIC
        --------------
        if callable meth/func - call and return result.
        else - return source.

        GOAL
        ----
        get common expected for any python code result - simple calculate or raise!
        because of get_result_or_exx is not enough!

        CREATED SPECIALLY FOR
        ---------------------
        GetattrPrefixInst
        """
        args = args or []
        kwargs = kwargs or {}

        if TypeChecker.check__callable_func_meth_inst(source):
            result = source(*args, **kwargs)
        else:
            result = source
        return result

    # -----------------------------------------------------------------------------------------------------------------
    @classmethod
    def get_result_or_exx(
            cls,
            source: TYPE__VALID_SOURCE,
            args: TYPE__VALID_ARGS = None,
            kwargs: TYPE__VALID_KWARGS = None,
    ) -> Any | Exception:
        """
        GOAL
        ----
        same as get_result_or_raise but
        attempt to simplify result by not using try-sentence.
        so if get raise in get_result_or_raise - return Exx object

        USEFUL IDEA
        -----------
        1. in gui when its enough to get str() on result and see the result
        """
        try:
            result = cls.get_result_or_raise(source=source, args=args, kwargs=kwargs)
        except Exception as exx:
            result = exx
        return result

    # -----------------------------------------------------------------------------------------------------------------
    @classmethod
    def get_result_bool(
            cls,
            source: TYPE__VALID_SOURCE,
            args: TYPE__VALID_ARGS = None,
            kwargs: TYPE__VALID_KWARGS = None,
    ) -> bool:
        """
        GOAL
        ----
        same as get_result_or_exx but
        apply bool func on result

        ability to get bool result with meanings:
            - methods/funcs must be called
                assert get_bool(LAMBDA_TRUE) is True
                assert get_bool(LAMBDA_NONE) is False

            - Exceptions assumed as False
                assert get_bool(Exception) is False
                assert get_bool(Exception("FAIL")) is False
                assert get_bool(LAMBDA_EXX) is False

            - for other values get classic bool()
                assert get_bool(None) is False
                assert get_bool([]) is False
                assert get_bool([None, ]) is True

                assert get_bool(LAMBDA_LIST) is False
                assert get_bool(LAMBDA_LIST, [1, ]) is True

            - if on bool() exception raised - return False!
                assert get_bool(ClsBoolExx()) is False

        CREATED SPECIALLY FOR
        ---------------------
        funcs.Valid.skip_link or else value/func assumed as bool result
        """
        try:
            result = cls.get_result_or_raise(source, args, kwargs)
            if TypeChecker.check__exception(result):
                return False
            return bool(result)
        except:
            return False

    # =================================================================================================================
    @classmethod
    def compare_doublesided_or_exx(cls, obj1: Any, obj2: Any, return_bool: bool = None) -> bool | Exception:
        """
        GOAL
        ----
        just a direct comparing code like
            self.validate_last = self.value_last == self.VALIDATE_LINK or self.VALIDATE_LINK == self.value_last
        will not work correctly

        if any result is True - return True.
        if at least one false - return False
        if both exx - return first exx  # todo: deside return False in here!

        CREATED SPECIALLY FOR
        ---------------------
        manipulate objects which have special methods for __cmp__
        for cases when we can switch places

        BEST USAGE
        ----------
            class ClsEq:
                def __init__(self, val):
                    self.VAL = val

                def __eq__(self, other):
                    return other == self.VAL

            assert ClsEq(1) == 1
            assert 1 == ClsEq(1)

            assert compare_doublesided(1, Cls(1)) is True
            assert compare_doublesided(Cls(1), 1) is True

        example above is not clear! cause of comparison works ok if any of object has __eq__() meth even on second place!
        but i think in one case i get Exx and with switching i get correct result!!! (maybe fake! need explore!)
        """
        try:
            result12 = obj1 == obj2
            if result12:
                return True
        except Exception as exx:
            result12 = exx

        try:
            result21 = obj2 == obj1
            if result21:
                return True
        except Exception as exx:
            result21 = exx

        try:
            result3 = obj2 is obj1
            if result3:
                return True
        except Exception as exx:
            result3 = exx
            pass

        if False in [result12, result21] or return_bool:
            return False
        else:
            return result12

    @classmethod
    def compare_doublesided__bool(cls, obj1: Any, obj2: Any) -> bool:
        """
        same as compare_doublesided_or_exx but
        in case of Exx - return False

        CREATED SPECIALLY FOR
        ---------------------
        Valid.value_validate
        """
        return cls.compare_doublesided_or_exx(obj1, obj2, return_bool=True)

    @classmethod
    def compare_doublesided__reverse(cls, obj1: Any, obj2: Any) -> bool:
        """
        just reverse result for compare_doublesided__bool
        so never get Exx, only bool
        """
        return cls.compare_doublesided__bool(obj1, obj2) is not True

    # =================================================================================================================
    @staticmethod
    def ltgt(source: Any, low: Any | None = None, high: Any | None = None) -> bool | Exception:
        """
        NOTE
        ----
        1. important to keep source at first place!
        """
        result = True
        if low is not None:
            result &= source > low
        if high is not None:
            result &= source < high
        return result

    @staticmethod
    def ltge(source: Any, low: Any | None = None, high: Any | None = None) -> bool | Exception:
        result = True
        if low is not None:
            result &= source > low
        if high is not None:
            result &= source <= high
        return result

    @staticmethod
    def legt(source: Any, low: Any | None = None, high: Any | None = None) -> bool | Exception:
        result = True
        if low is not None:
            result &= source >= low
        if high is not None:
            result &= source < high
        return result

    @staticmethod
    def lege(source: Any, low: Any | None = None, high: Any | None = None) -> bool | Exception:
        result = True
        if low is not None:
            result &= source >= low
        if high is not None:
            result &= source <= high
        return result


# =====================================================================================================================



from typing import *
import pathlib

from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *


# =====================================================================================================================
# TODO: add COMBO with context-manu for add+delete+lastValue
# TODO: create blank Objects to change state!!! GREEn/RED/Yellow/bare


# =====================================================================================================================
# original names is not enough soft and comfortable for eyes!
# try to use ROUND NUMBERS!!! start use 0/100/200/255!!!
COLOR_TUPLE_RGB__LIGHT_RED = (255, 200, 200)
COLOR_TUPLE_RGB__LIGHT_GREEN = (200, 255, 200)
COLOR_TUPLE_RGB__LIGHT_BLUE = (200, 200, 255)
COLOR_TUPLE_RGB__LIGHT_YELLOW = (255, 255, 100)
COLOR_TUPLE_RGB__LIGHT_GREY_240 = (240, 240, 240)   # #f2f2f2
COLOR_TUPLE_RGB__LIGHT_GREY_220 = (220, 220, 220)
COLOR_TUPLE_RGB__LIGHT_GREY_150 = (150, 150, 150)
COLOR_TUPLE_RGB__WHITE = (255, 255, 255)

MARGINS_0000 = QMargins(0, 0, 0, 0)
MARGINS_5555 = QMargins(5, 5, 5, 5)
MARGINS_9595 = QMargins(9, 5, 9, 5)

ALIGNMENT_T = Qt.Alignment(Qt.AlignTop)
ALIGNMENT_LT = Qt.Alignment(Qt.AlignLeft | Qt.AlignTop)
ALIGNMENT_CENTER_HV = Qt.Alignment(Qt.AlignHCenter | Qt.AlignVCenter)


# ICONS ===============================================================================================================
class QIcon_Set:
    _path: pathlib.Path = pathlib.Path().parent.joinpath('images')

    def __init__(self):
        # JPG
        self.logo_square_200: QIcon = QIcon(str(self._path.joinpath('logo_square_200.jpg')))

        # PNG
        self.load_settings: QIcon = QIcon(str(self._path.joinpath('load_settings.png')))
        self.init: QIcon = QIcon(str(self._path.joinpath('init.png')))
        self.ok: QIcon = QIcon(str(self._path.joinpath('ok.png')))
        self.error: QIcon = QIcon(str(self._path.joinpath('error.png')))
        self.save: QIcon = QIcon(str(self._path.joinpath('save.png')))

        # SVG
        self.arrow_right: QIcon = QIcon(str(self._path.joinpath('arrow_right.svg')))
        self.folder2_open: QIcon = QIcon(str(self._path.joinpath('folder2_open.svg')))
        self.gear: QIcon = QIcon(str(self._path.joinpath('gear.svg')))


# COLOR ===============================================================================================================
class QWidget_ColorChange(QWidget):
    def _color_set_rgb(self, color=None):
        if color:
            self.setStyleSheet(f"background: rgb{color}")
        else:
            self.setStyleSheet(f"")

    def color_clear(self):
        self._color_set_rgb(None)

    def color_set__light_green(self):
        self._color_set_rgb(COLOR_TUPLE_RGB__LIGHT_GREEN)

    def color_set__light_red(self):
        self._color_set_rgb(COLOR_TUPLE_RGB__LIGHT_RED)

    def color_set__light_yellow(self):
        self._color_set_rgb(COLOR_TUPLE_RGB__LIGHT_YELLOW)


class QPushButton(QPushButton, QWidget_ColorChange):
    pass


class QLineEdit(QLineEdit, QWidget_ColorChange):
    pass


class QComboBox(QComboBox, QWidget_ColorChange):
    pass


class QLabel(QLabel, QWidget_ColorChange):
    pass


# =====================================================================================================================
class QPushButton_Checkable(QPushButton):
    def __init__(self, *args, **kwargs):
        self.__text_toggle: list = ["PushDown", "Release"]  # dont move to class method!!! will not work correct!

        # make available init with list!!! -------------------
        args = list(args)
        for index, arg in enumerate(args):
            if isinstance(arg, str):
                self.__text_toggle[0] = self.__text_toggle[1] = arg
                break
            elif isinstance(arg, list) and len(arg) == 2:
                self.__text_toggle = arg
                args[index] = arg[0]
                break
        super().__init__(*args, **kwargs)

        self.__text_toggle_state()

        # work --------------------------------
        self.setCheckable(True)
        self.toggled.connect(self.__text_toggle_state)

    def __text_toggle_state(self, state=None):
        if state is None:
            state = int(self.isChecked())

        self.setText(self.__text_toggle[state])

    def setText(self, text):
        index = int(self.isChecked())
        self.__text_toggle[index] = text
        super().setText(text)


# TESTS ===============================================================================================================
class __TestExample(QWidget):
    def __init__(self):
        super().__init__()

        self.btn1 = QPushButton_Checkable("1")
        self.btn2 = QPushButton_Checkable(["21", "22"])
        self.btn3 = QPushButton_Checkable()

        layout = QVBoxLayout()
        layout.addWidget(self.btn1)
        layout.addWidget(self.btn2)
        layout.addWidget(self.btn3)
        self.setLayout(layout)
        self.show()

    def print_msg(self):
        print()
        print(self.btn1.isChecked())
        print(self.btn2.isChecked())
        print(self.btn3.isChecked())


# =====================================================================================================================
if __name__ == "__main__":
    app = QApplication([])
    wgt = __TestExample()
    print(app.exec_())


# =====================================================================================================================



# ===================================================================
from base_aux.privates import *
import pathlib


# ===================================================================
# by instance attr
# {"AUTH": {"NAME": "MyName", "PWD": "MyPwd"}}

class Cls:
    data = PrivateAuthJson(_section="AUTH")
    def connect(self):
        name = self.data.NAME
        name = self.data.NamE     # case insensitive

# like dict key on instance
# {"AUTH": {"NAME": "MyName", "PWD": "MyPwd"}}


class Cls:
    data = PrivateAuthJson(_section="AUTH")
    def connect(self):
        name = self.data["NAME"]
        name = self.data["NamE"]   # case insensitive

# ===================================================================
### use annotations for your param names (best practice!)
# when instantiating if it will not get loaded these exact params from your private sources - RAISE!  
# but you could not use it and however keep access to all existed params in used section!
# {"AUTH": {"NAME": "MyName", "PWD": "MyPwd"}}

class MyPrivateJson(PrivateJson):
    NAME: str
    PWD: str

name = MyPrivateJson().NAME


# ===================================================================
# in example above you could simply use existed classes
name = PrivateAuthJson().NAME


# ===================================================================
### 1. Env

class Cls:
   user = PrivateEnv["NAME"]
   user = PrivateEnv.NAME


# ===================================================================
### 2. IniFile
# Use different sections
class Cls:
   user = PrivateIni(_section="CustomSection").NAME


# ===================================================================
# Change full settings
class CustomIniValues(PrivateIni):
   DIRPATH = "new/path/"
   DIRPATH = pathlib.Path("new/path/")
   FILENAME = "my.ini"
   SECTION = "CustomSection"

class Cls:
   user = CustomIniValues.NAME

# ===================================================================
# Without creating new class
class Cls:
   pv1 = PrivateIni(_filename="otherFilename").pv1


# ===================================================================
### 3. JsonFile
# {"AUTH": {"NAME": "MyName", "PWD": "MyPwd"}}
class MyPrivateJson(PrivateJson):
    SECTION = "AUTH"
    NAME: str
    PWD: str

class Cls:
    data = MyPrivateJson()
    def connect(self):
        name = self.data.NAME

# ===================================================================
# use already created templates (PrivateAuthJson/PrivateTgBotAddressJson) for standard attributes
# {"AUTH": {"NAME": "MyName", "PWD": "MyPwd"}}

class Cls:
    data = PrivateAuthJson(_section="AUTH")
    def connect(self):
        name = self.data.NAME

# ===================================================================
### 4. Auto  
# you can use universal class  
# it will trying get all your annotated params from one source of Json/Ini/Env (in exact order)  
# in this case you cant use FileName and must use annotations!

# {"AUTH": {"NAME": "MyName", "PWD": "MyPwd"}}

class MyPrivate(PrivateAuto):
    SECTION = "AUTH"
    NAME: str
    PWD: str

name = MyPrivate().NAME

# ===================================================================



from typing import *
from base_aux.testplans import DutBase
from base_aux.buses import *


# =====================================================================================================================
class Device(SerialClient_FirstFree_AnswerValid, DutBase):
    LOG_ENABLE = True
    RAISE_CONNECT = False
    BAUDRATE = 115200
    EOL__SEND = b"\n"

    REWRITEIF_READNOANSWER = 0
    REWRITEIF_NOVALID = 0

    # MODEL INFO --------------------------------
    __sn_start: str = "SN"
    NAME: str = "PSU"
    DESCRIPTION: str = "Power Supply Unit"

    @property
    def SN(self) -> str:
        return f"{self.__sn_start}_{self.INDEX+1}"

    @SN.setter
    def SN(self, value: Any) -> None:
        self.__sn_start = str(value).upper()
    # MODEL INFO --------------------------------

    @property
    def DEV_FOUND(self) -> bool:
        return True
        return self.address_check__resolved()

    @property
    def PREFIX(self) -> str:
        return f"PTB:{self.INDEX+1:02d}:"

    def address__validate(self) -> bool:
        return True
        result = (
                self.write_read__last_validate("get name", f"PTB", prefix="*:")
                and
                self.write_read__last_validate("get addr", [f"{self.INDEX+1}", f"0{self.INDEX+1}"], prefix="*:")
                # and
                # self.write_read__last_validate("get prsnt", "0")
        )
        return result

    def connect__validate(self) -> bool:
        return True
        result = (
                self.write_read__last_validate("get prsnt", "0")
        )
        return result

    def __init__(self, index: int, **kwargs):
        if index is not None:
            self.INDEX = index
        super().__init__(**kwargs)

    @property
    def VALUE(self) -> bool:
        return self.INDEX % 2 == 0

    def connect(self):
        return True


# =====================================================================================================================
if __name__ == "__main__":
    pass

    # emu = Ptb_Emulator()
    # emu.start()
    # emu.wait()

    dev = Device(1)
    print(f"{dev.connect()=}")
    print(f"{dev.ADDRESS=}")

    # dev.write_read__last("get sn")
    # dev.write_read__last("get fru")
    # dev.write_read__last("test sc12s")
    # dev.write_read__last("test ld12s")
    # dev.write_read__last("test gnd")
    # dev.write_read__last("test")
    # dev.write_read__last("get status")
    dev.write_read__last("get vin")


# =====================================================================================================================



from typing import *
import numpy as np


# =====================================================================================================================
def array_2d_get_compact_str(
        array: np.ndarray,
        interpreter: Optional[Dict[Any, Any]] = None,
        separate_rows: Optional[int] = None,
        wrap: Optional[bool] = None,
        use_rows_num: Optional[bool] = None
) -> str:
    """

    :param array:
    :param interpreter: dictionary to change some elements
    :param separate_rows: add blank line on step
    :param wrap: add additional strings before and after data
    :param use_rows_num: add row num in
    :return:
    """
    interpreter = interpreter or {}
    count_rows = len(array)
    count_columns = len(array[0])
    row_pos = 0
    result: str = ""

    # tab_row_nums = (count_rows +1)//10 + 1
    # if (count_rows+1)%10 > 0:
    #     tab_row_nums += 1

    tab_row_nums = 4

    if wrap:
        if use_rows_num:
            result += " " * tab_row_nums
        result += "=" * count_columns + "\n"

    for row in array:
        row_pos += 1
        if separate_rows and row_pos > 1 and (row_pos - 1) % separate_rows == 0:
            result += f"\n"

        if use_rows_num:
            result += "{:{width}}".format(str(row_pos), width=tab_row_nums)
        for value in row:
            replaced = interpreter.get(value)
            if replaced is not None:
                value = replaced
            result += f"{value}"

        if row_pos != count_rows:
            result += f"\n"

    if wrap:
        result += "\n"
        if use_rows_num:
            result += " " * tab_row_nums
        result += "=" * count_columns

    return result


# =====================================================================================================================



# =====================================================================================================================
# VERSION = (0, 0, 1)   # use import EXACT_OBJECTS! not *
#   from .main import *                 # INcorrect
#   from .main import EXACT_OBJECTS     # CORRECT


# =====================================================================================================================
# TEMPLATE
# from .main import (
#     # BASE
#     EXACT_OBJECTS,
#
#     # AUX
#
#     # TYPES
#
#     # EXX
# )
# ---------------------------------------------------------------------------------------------------------------------
from .strings import (
    # BASE
    ReqCheckStr_Base,
    ReqCheckStr_Os,
    ReqCheckStr_Arch,

    # AUX
    _GetattrClassmethod_Meta,

    # TYPES
    TYPE__VALUES,
    TYPE__RESULT_BOOL,
    TYPE__RESULT_RAISE,

    # EXX
    Exx_RequirementCantGetActualValue,
    Exx_Requirement,
)
from .pkgs import (
    # BASE
    Packages,
    CmdPattern,

    # AUX
    PATTERNS_IMPORT,
    PATTERN_IMPORT__MULTY_COMMA,
    PATTERN_IMPORT__MULTY_COMMA_BRACKETS,
    PATTERN_IMPORT__FROM,
    # TYPES
    # EXX
)
from .versions import (
    # BASE
    Version,
    ReqCheckVersion,
    ReqCheckVersion_Python,

    # AUX
    VersionBlock,

    # TYPES
    TYPE__VERSION_ELEMENT,
    TYPE__VERSION_ELEMENTS,
    TYPE__SOURCE_BLOCKS,

    TYPE__VERSION_BLOCKS,
    TYPE__SOURCE_VERSION,

    # EXX
    Exx_VersionIncompatible,
    Exx_VersionIncompatibleBlock,
    Exx_VersionIncompatibleCheck,
)


# =====================================================================================================================



from . import *

from typing import *
import re


# =====================================================================================================================
class LineParsed:
    """
    ALL RESULTS IN LOWERCASE! (EXCEPT ORIGINAL ORIGINAL!)
    """
    # REAL STATE --------------
    ORIGINAL: str
    PREFIX: str
    CMD: str
    ARGS: list[str]
    KWARGS: dict[str, str]

    # AUX ---------------------
    _PREFIX_EXPECTED: str

    def __init__(self, line: str, _prefix_expected: Optional[str] = None):
        line = str(line)

        # INIT ----------------
        self.ORIGINAL = line
        self.PREFIX = ""
        self.CMD = ""
        self.ARGS = []
        self.KWARGS = {}

        line_lower = line.lower()

        # PREFIX ----------------
        _prefix_expected = _prefix_expected or ""
        _prefix_expected = _prefix_expected.lower()
        self._PREFIX_EXPECTED = _prefix_expected
        if _prefix_expected and line_lower.startswith(_prefix_expected):
            self.PREFIX = _prefix_expected
            line_lower = line_lower.replace(_prefix_expected, "", 1)

        # BLANK ----------------------
        if not line_lower:
            return
        line_lower = re.sub(r"\s*=+\s*", "=", line_lower)
        line_parts = line_lower.split()
        if not line_parts:
            return

        # ARGS/KWARGS ----------------
        for part in line_parts:
            if "=" not in part:
                self.ARGS.append(part)
            else:
                part__key_value = part.split("=")
                self.KWARGS.update(dict([part__key_value, ]))

        # CMD ------------------------
        if self.ARGS:
            self.CMD = self.ARGS[0]
            self.ARGS = self.ARGS[1:]

        pass

    def ARGS_count(self) -> int:
        return len(self.ARGS)

    def KWARGS_count(self) -> int:
        return len(self.KWARGS)


# =====================================================================================================================



from typing import *
from base_aux.funcs import IterAux, Explicit
from base_aux.classes.attr_3_dict_dots_1_simple import DictDots


# =====================================================================================================================
class DictDotsFull(DictDots):
    """
    TODO: do this perfect idea!

    smth like BS4 syntax but for dict/list
    1. use full access WR work with any level
    2. dict/list - ok! tuple - as generic final!
    3. keep source link! not just a simple copy (like in DictDots)!

    WHY NOT - DictDots
    ------------------
    it was a simple way to work with mainly first level! and only with dict!
    it seems it is obviously need use with lists same!
    """


# =====================================================================================================================



import sys
from typing import *
import re

from base_aux.objects import *
from base_aux.funcs import *
from base_aux.classes import *


# =====================================================================================================================
# TODO:


# =====================================================================================================================
# EXPLORE VARIANTS ----------------------------------------------
# from packaging import version
# ObjectInfo(version.parse(str((1,2,3)))).print()

# result = version.parse("2.3.1") < version.parse("10.1.2")

# ObjectInfo(version.parse("1.2.3")).print()
# print(result)
# print()
#
# from pkg_resources import parse_version         # DEPRECATED!!!
# parse_version("1.9.a.dev") == parse_version("1.9a0dev")
#

# import sys
# print(sys.winver)
# print(sys.version_info)
# print(tuple(sys.version_info))
#
# result = sys.version_info > (2, 7)
# print(result)


# =====================================================================================================================
class Exx_VersionIncompatibleBlock(Exception):
    """
    """


class Exx_VersionIncompatible(Exception):
    """
    """


class Exx_VersionIncompatibleCheck(Exception):
    """
    """


# =====================================================================================================================
TYPE__VERSION_ELEMENT = Union[str, int]
TYPE__VERSION_ELEMENTS = tuple[TYPE__VERSION_ELEMENT, ...]
TYPE__SOURCE_BLOCKS = Union[str, int, list, tuple, Any, 'VersionBlock']


class VersionBlock(CmpInst):
    """
    this is exact block in version string separated by dots!!!

    PATTERN for blocks
    ------------------
        block1.block2.block3

    EXAMPLES for block
    ------------------
        1rc2
        1-rc2
        1 rc 2

    RULES
    -----
    1.
    """
    _SOURCE: Any
    ELEMENTS: TYPE__VERSION_ELEMENTS

    PATTERN_CLEAR = r"[\"' -]*"
    PATTERN_VALIDATE_SOURCE_NEGATIVE = r"\d+[^0-9a-z]+\d+"
    PATTERN_VALIDATE_CLEANED = r"(\d|[a-z])+"
    PATTERN_ITERATE = r"\d+|[a-z]+"

    def __init__(self, source: TYPE__SOURCE_BLOCKS):
        self._SOURCE = source
        if not self._validate_source(source):
            raise Exx_VersionIncompatibleBlock()

        string = self._prepare_string(source)
        if not self._validate_string(string):
            raise Exx_VersionIncompatibleBlock()

        self.ELEMENTS = self._parse_elements(string)

    @classmethod
    def _validate_source(cls, source: TYPE__SOURCE_BLOCKS) -> bool:
        source = str(source).lower()
        match = re.search(cls.PATTERN_VALIDATE_SOURCE_NEGATIVE, source)
        return not bool(match)

    @classmethod
    def _prepare_string(cls, source: TYPE__SOURCE_BLOCKS) -> str:
        if isinstance(source, (list, tuple)):
            result = "".join([str(item) for item in source])
        else:
            result = str(source)

        # FINISH -------------------------------
        result = re.sub(cls.PATTERN_CLEAR, "", result)
        result = result.lower()
        result = result.strip()
        return result

    @classmethod
    def _validate_string(cls, string: str) -> bool:
        if not isinstance(string, str):
            return False
        match = re.fullmatch(cls.PATTERN_VALIDATE_CLEANED, string)
        return bool(match)

    @classmethod
    def _parse_elements(cls, string: str) -> TYPE__VERSION_ELEMENTS:
        if not isinstance(string, str):
            return ()

        result_list = []
        for element in re.findall(cls.PATTERN_ITERATE, string):
            try:
                element = int(element)
            except:
                pass
            result_list.append(element)

        return tuple(result_list)

    def __iter__(self):
        yield from self.ELEMENTS

    def __len__(self) -> int:
        return len(self.ELEMENTS)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self})"

    def __str__(self) -> str:
        return "".join(str(item) for item in self.ELEMENTS)

    @property
    def STRING(self) -> str:
        return str(self)

    # CMP -------------------------------------------------------------------------------------------------------------
    def __cmp__(self, other) -> int | NoReturn:
        other = self.__class__(other)

        # equel ----------------------
        if str(self) == str(other):
            return 0

        # by elements ----------------
        for elem_1, elem_2 in zip(self, other):
            if elem_1 == elem_2:
                continue

            if isinstance(elem_1, int):
                if isinstance(elem_2, int):
                    return elem_1 - elem_2
                else:
                    return 1
            else:
                if isinstance(elem_2, int):
                    return -1
                else:
                    return int(elem_1 > elem_2) or -1

        # final - longest ------------
        return int(len(self) > len(other)) or -1


# =====================================================================================================================
pass    # -------------------------------------------------------------------------------------------------------------
pass    # -------------------------------------------------------------------------------------------------------------
pass    # -------------------------------------------------------------------------------------------------------------
pass    # -------------------------------------------------------------------------------------------------------------
pass    # -------------------------------------------------------------------------------------------------------------
pass    # -------------------------------------------------------------------------------------------------------------
pass    # -------------------------------------------------------------------------------------------------------------
pass    # -------------------------------------------------------------------------------------------------------------

TYPE__VERSION_BLOCKS = tuple[VersionBlock, ...]
TYPE__SOURCE_VERSION = Union[VersionBlock, tuple[VersionBlock], 'Version', Any]


class PatternsVer:
    # VERSION_TUPLE = r"\((\d+\.+(\w+\.?)+)\)"
    # VERSION_LIST = r"\[(\d+\.+(\w+\.?)+)\]"
    VERSION_IN_BRACKETS: list = [r"\((.*)\)", r"\[(.*)\]"]  # get first bracket!!!
    VALIDATE_BRACKETS_NEGATIVE: list = [r"[^\[].*\]", r"\[.*[^\]]",   r"[^\(].*\)", r"\(.*[^\)]"]


# =====================================================================================================================
class Version(CmpInst):
    """
    :ivar _SOURCE: try to pass parsed value! it will try to self-parse in _prepare_string, but make it ensured on your own!
    """
    _SOURCE: Any
    BLOCKS: TYPE__VERSION_BLOCKS = ()

    MIN_BLOCKS_COUNT: int = 1
    RAISE: bool = True

    def __init__(self, source: Any, min_blocks_count: int = None):
        if min_blocks_count is not None:
            self.MIN_BLOCKS_COUNT = min_blocks_count

        self._SOURCE = source
        string = self._prepare_string(source)
        self.BLOCKS = self._parse_blocks(string)
        if not self.check_blocks_enough() and self.RAISE:
            raise Exx_VersionIncompatible()

    def check_blocks_enough(self, count: int = None) -> bool:
        if count is None:
            count = self.MIN_BLOCKS_COUNT
        return len(self.BLOCKS) >= count

    # -----------------------------------------------------------------------------------------------------------------
    @classmethod
    def _prepare_string(cls, source: Any) -> str:
        """
        ONLY PREPARE STRING FOR CORRECT SPLITTING BLOCKS - parsing blocks would inside VersionBlock
        """
        if isinstance(source, (list, tuple)):
            result = ".".join([str(item) for item in source])
        else:
            result = str(source)

        result = result.lower()

        # CUT ---------
        for pattern in PatternsVer.VERSION_IN_BRACKETS:
            match = re.search(pattern, result)
            if match:
                result = match[1]
                break

        if "," in result and "." in result:
            raise Exx_VersionIncompatible()

        for pattern in PatternsVer.VALIDATE_BRACKETS_NEGATIVE:
            if re.search(pattern, result):
                raise Exx_VersionIncompatible()

        result = re.sub(r"\A\D+", "", result)   # ver/version
        result = re.sub(r",+", ".", result)
        result = re.sub(r"\.+", ".", result)
        result = result.strip(".")

        return result

    @staticmethod
    def _parse_blocks(source: str) -> TYPE__VERSION_BLOCKS:
        blocks_list__str = source.split(".")

        # RESULT -----------
        result = []
        for item in blocks_list__str:
            if not item:
                continue

            block = VersionBlock(item)
            result.append(block)

        return tuple(result)

    # -----------------------------------------------------------------------------------------------------------------
    def __len__(self) -> int:
        return len(self.BLOCKS)

    def __getitem__(self, item: int) -> VersionBlock | None:
        try:
            return self.BLOCKS[item]
        except:
            return

    def __iter__(self):
        yield from self.BLOCKS

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self})"

    def __str__(self):
        return ".".join([str(block) for block in self.BLOCKS])

    @property
    def STRING(self) -> str:
        return str(self)

    # -----------------------------------------------------------------------------------------------------------------
    @property
    def major(self) -> VersionBlock | None:
        return self[0]

    @property
    def minor(self) -> VersionBlock | None:
        return self[1]

    @property
    def micro(self) -> VersionBlock | None:
        return self[2]

    # CMP -------------------------------------------------------------------------------------------------------------
    def __cmp__(self, other: TYPE__SOURCE_VERSION) -> int | NoReturn:
        other = self.__class__(other)

        # equel ----------------------
        if str(self) == str(other):
            return 0

        # by elements ----------------
        for block_1, block_2 in zip(self, other):
            if block_1 == block_2:
                continue
            else:
                return int(block_1 > block_2) or -1

        # final - longest ------------
        return int(len(self) > len(other)) or -1


# =====================================================================================================================
class ReqCheckVersion(GetattrPrefixInst_RaiseIf):
    SOURCE: Union[Any, Callable[..., Any]] = sys.version.split()[0]

    # -----------------------------------------------------------------------------------------------------------------
    def __init__(self, source: Any | Callable | None = None):
        if source is not None:
            self.SOURCE = source

    @property
    def ACTUAL(self) -> Version:
        if TypeChecker.check__callable_func_meth_inst(self.SOURCE):
            value = self.SOURCE()
        else:
            value = self.SOURCE

        return Version(value)

    # ---------------------------------------
    def check_eq(self, target: Any):
        return self.ACTUAL == target

    def check_ne(self, target: Any):
        return self.ACTUAL != target

    # ---------------------------------------
    def check_le(self, target: Any):
        return self.ACTUAL <= target

    def check_lt(self, target: Any):
        return self.ACTUAL < target

    # ---------------------------------------
    def check_ge(self, target: Any):
        return self.ACTUAL >= target

    def check_gt(self, target: Any):
        return self.ACTUAL > target


# ---------------------------------------------------------------------------------------------------------------------
class ReqCheckVersion_Python(ReqCheckVersion):
    """
    check version of python interpreter.

    USAGE
    -----
    ReqCheckVersion_Python().raise_if_not__check_ge("2")
    ReqCheckVersion_Python().raise_if_not__check_ge("3.11")
    ReqCheckVersion_Python().raise_if_not__check_ge("3.11rc1", _comment="need Python GRATER EQUAL")
    """
    SOURCE = sys.version.split()[0]

    raise_if__check_eq: Callable[..., NoReturn | None]
    raise_if_not__check_eq: Callable[..., NoReturn | None]

    raise_if__check_ne: Callable[..., NoReturn | None]
    raise_if_not__check_ne: Callable[..., NoReturn | None]

    raise_if__check_le: Callable[..., NoReturn | None]
    raise_if_not__check_le: Callable[..., NoReturn | None]

    raise_if__check_lt: Callable[..., NoReturn | None]
    raise_if_not__check_lt: Callable[..., NoReturn | None]

    raise_if__check_ge: Callable[..., NoReturn | None]
    raise_if_not__check_ge: Callable[..., NoReturn | None]

    raise_if__check_gt: Callable[..., NoReturn | None]
    raise_if_not__check_gt: Callable[..., NoReturn | None]


# =====================================================================================================================



# =====================================================================================================================
# VERSION = (0, 0, 1)   # use import EXACT_OBJECTS! not *
#   from .main import *                 # INcorrect
#   from .main import EXACT_OBJECTS     # CORRECT


# =====================================================================================================================
from .info import (
    # BASE
    ObjectInfo,
    # AUX
    ItemInternal,
    ObjectState,
    # TYPES
    # EXX
)
from .obj_types import (
    # BASE
    TypeChecker,
    # AUX
    # TYPES
    TYPE__NONE,
    TYPE__FUNCTION,
    TYPE__METHOD,
    # EXX
)

# ---------------------------------------------------------------------------------------------------------------------
from .primitives import (
    # BASE
    GEN_COMPR,

    FUNC,
    FUNC_NONE,
    FUNC_TRUE,
    FUNC_FALSE,
    FUNC_ALL,
    FUNC_ANY,
    FUNC_LIST_DIRECT,
    FUNC_LIST_VALUES,
    FUNC_DICT,
    FUNC_EXX,
    FUNC_GEN,

    LAMBDA,
    LAMBDA_NONE,
    LAMBDA_TRUE,
    LAMBDA_FALSE,
    LAMBDA_ALL,
    LAMBDA_ANY,
    LAMBDA_LIST_DIRECT,
    LAMBDA_LIST_VALUES,
    LAMBDA_DICT,
    LAMBDA_EXX,
    LAMBDA_GEN,

    Exx,            INST_EXX,

    ClsInt,
    ClsFloat,
    ClsStr,
    ClsList,
    ClsSet,
    ClsDict,

    Cls,            INST,
    ClsEmpty,       INST_EMPTY,

    ClsInitArgsKwargs,
    ClsInitExx,

    ClsCall,        INST_CALL,
    ClsCallNone,    INST_CALL_NONE,
    ClsCallTrue,    INST_CALL_TRUE,
    ClsCallFalse,   INST_CALL_FALSE,
    ClsCallExx,     INST_CALL_EXX,

    ClsBoolTrue,    INST_BOOL_TRUE,
    ClsBoolFalse,   INST_BOOL_FALSE,
    ClsBoolExx,     INST_BOOL_EXX,

    ClsIterYield,   INST_ITER_YIELD,
    ClsGen,         INST_GEN,

    ClsEq,          INST_EQ,
    ClsEqTrue,      INST_EQ_TRUE,
    ClsEqFalse,     INST_EQ_FALSE,
    ClsEqExx,       INST_EQ_EXX,

    ClsFullTypes,   INST_FULL_TYPES,

    CALLABLE_LAMBDA,
    CALLABLE_FUNC,

    CALLABLE_CLS,
    CALLABLE_INST,

    CALLABLE_METH_CLS,
    CALLABLE_METH_CLS_CLASSMETHOD,
    CALLABLE_METH_CLS_STATICMETHOD,
    CALLABLE_METH_CLS_PROPERTY,
    CALLABLE_METH_CLS_PROPERTY_CLASSMETHOD,

    CALLABLE_METH_INST,
    CALLABLE_METH_INST_CLASSMETHOD,
    CALLABLE_METH_INST_STATICMETHOD,
    CALLABLE_METH_INST_PROPERTY,
    CALLABLE_METH_INST_PROPERTY_CLASSMETHOD,
    # AUX
    # TYPES
    # EXX
)

# =====================================================================================================================



from typing import *
from .url_tag import *


# =====================================================================================================================
class Monitor_DonorSvetofor(MonitorUrlTag):
    """
    MONITOR donor svetofor and alert when BloodCenter need your blood group!

    # STRUCTURE to find -------------------------------------
<table class="donor-svetofor-restyle">
    <tbody>
        <tr>
            <th colspan="2">O (I)</th>
            <th colspan="2">A (II)</th>
            <th colspan="2">B (III)</th>
            <th colspan="2">AB (IV)</th>
        </tr>
        <tr>
            <td class="green">Rh +</td>
            <td class="green">Rh –</td>
            <td class="green">Rh +</td>
            <td class="yellow">Rh –</td>
            <td class="green">Rh +</td>
            <td class="yellow">Rh –</td>
            <td class="red">Rh +</td>
            <td class="red">Rh –</td>
        </tr>
    </tbody>
</table>
    """
    # KEEP FIRST!
    DONOR_BLOOD_GROUP: int = 3
    DONOR_BLOOD_RH: str = "+"

    # MAIN -------------------------------
    URL = "https://donor.mos.ru/donoru/donorskij-svetofor/"
    TAG_CHAINS = [
        TagAddressChain("table", {"class": "donor-svetofor-restyle"}, None, 0),
        TagAddressChain("td", {}, f"Rh {DONOR_BLOOD_RH}", DONOR_BLOOD_GROUP - 1),
    ]
    TAG_GET_ATTR = "class"


# =====================================================================================================================
class Monitor_CbrKeyRate(MonitorUrlTag):
    """
    MONITOR CentralBankRussia KeyRate

    # STRUCTURE to find -------------------------------------
<div class="table-wrapper">
  <div class="table-caption gray">% годовых</div>
  <div class="table">
    <table class="data">
      <tr>
        <th>Дата</th>
        <th>Ставка</th>
      </tr>
      <tr>
        <td>17.08.2023</td>
        <td>12,00</td>
      </tr>
      <tr>
        <td>16.08.2023</td>
        <td>12,00</td>
      </tr>
      <tr>
        <td>15.08.2023</td>
        <td>12,00</td>
      </tr>
      <tr>
        <td>14.08.2023</td>
        <td>8,50</td>
      </tr>
      <tr>
        <td>11.08.2023</td>
        <td>8,50</td>
      </tr>
      <tr>
        <td>10.08.2023</td>
        <td>8,50</td>
      </tr>
    </table>
  </div>
  <div class="table-caption">
    <p>
	  Данные доступны с  17.09.2013 по 17.08.2023.
	  </p>
  </div>
</div>
    """
    # OVERWRITING NEXT -------------------------------
    # KEEP FIRST!

    # OVERWRITTEN NOW -------------------------------
    URL = "https://cbr.ru/hd_base/KeyRate/"
    TAG_CHAINS = [
        TagAddressChain("div", {"class": "table-wrapper"}, None, 0),
        TagAddressChain("td", {}, None, 1),
    ]
    TAG_GET_ATTR = None


# =====================================================================================================================
class Monitor_ConquestS23_comments(MonitorUrlTag):
    """
    # STRUCTURE to find -------------------------------------
<div class="table-wrapper">
  <div class="table-caption gray">% годовых</div>
  <div class="table">
    <table class="data">
      <tr>
        <th>Дата</th>
        <th>Ставка</th>
      </tr>
      <tr>
        <td>17.08.2023</td>
        <td>12,00</td>
      </tr>
      <tr>
        <td>16.08.2023</td>
        <td>12,00</td>
      </tr>
      <tr>
        <td>15.08.2023</td>
        <td>12,00</td>
      </tr>
      <tr>
        <td>14.08.2023</td>
        <td>8,50</td>
      </tr>
      <tr>
        <td>11.08.2023</td>
        <td>8,50</td>
      </tr>
      <tr>
        <td>10.08.2023</td>
        <td>8,50</td>
      </tr>
    </table>
  </div>
  <div class="table-caption">
    <p>
	  Данные доступны с  17.09.2013 по 17.08.2023.
	  </p>
  </div>
</div>
    """
    # OVERWRITING NEXT -------------------------------
    # KEEP FIRST!

    # OVERWRITTEN NOW -------------------------------
    URL = "https://exgad.ru/products/conquest-s23"
    TAG_CHAINS = [
        TagAddressChain("div", {"class": "comments-tab__quatity"}, None, 0),
    ]
    TAG_GET_ATTR = None


# =====================================================================================================================
class Monitor_Sportmaster_AdidasSupernova2M(MonitorUrlTag):
    """
    # STRUCTURE to find -------------------------------------
    <span data-selenium="amount" class="sm-amount__value">13 699 ₽</span>
    """
    # TODO: need resolve StatusCode401
    # OVERWRITING NEXT -------------------------------
    # KEEP FIRST!

    # OVERWRITTEN NOW -------------------------------
    URL = "https://www.sportmaster.ru/product/29647730299/"
    TAG_CHAINS = [
        TagAddressChain("span", {"class": "sm-amount__value"}, None, 0),
    ]
    TAG_GET_ATTR = None


# =====================================================================================================================
class Monitor_AcraRaiting_GTLC(MonitorUrlTag):
    """
    NOT WORKING!
[Try send --------------------------------------------------------------------------------
[ALERT]Monitor_AcraRaiting_GTLC
LOST URL SSLError(MaxRetryError("HTTPSConnectionPool(host='www.acra-ratings.ru', port=443): Max retries exceeded with url: /ratings/issuers/50/ (Caused by SSLError(SSLCertVerificationError(1, '[SSL: CERTIFICATE_VERIFY_FAILED] certificate verify failed: unable to get local issuer certificate (_ssl.c:1002)')))"))[CRITICAL] empty self._source_data=''
2023.09.18 21:02:54
Try send] --------------------------------------------------------------------------------
[READY] send


    # STRUCTURE to find -------------------------------------
<div class="current-emit__actual-table">
    <div class="inner-table green" data-type="ratings">
        <div class="rating-head">
            <div class="item-info" data-type="scale">Шкала</div>
            <div class="item-info" data-type="rate">Рейтинг</div>
            <div class="item-info" data-type="pressRelease">Пресс-релиз</div>
        </div>
        <div class="rating-list">
            <div class="rating-item">
                <div class="item-info dedicated" data-type="scale">Национальная рейтинговая шкала</div>
                <div class="item-info dedicated" data-type="rate">
                    <div class="rating-widget" >
                        <div class="tooltip-ctrl" title="На пересмотре">
                            <svg class="svg status-icon">
                                <use xlink:href="/static/img/icons/symbol/sprite.svg#under-control"></use>
                            </svg>
                        </div>
                        AA-(RU)     На пересмотре (развивающийся)
                    </div>
                </div>
                <a
                    class="item-info dedicated 123123"
                    href="/press-releases/3966/"
                    data-type="pressRelease"
                >
                    18 май 2023
                    <img class="svg" src="/static/img/icons/search/link.svg">
                </a>
            </div>
        </div>
    </div>
    """
    TIMEOUT_REQUEST = 30
    URL = "https://www.acra-ratings.ru/ratings/issuers/50/"
    TAG_CHAINS = [
        TagAddressChain("div", {"class": "current-emit__actual-table"}, None, 0),
        TagAddressChain("div", {"class": "rating-list"}, None, 0),
        TagAddressChain("div", {"class": "rating-item"}, None, 0),
        TagAddressChain("div", {"class": "rating-widget"}, None, 0),
    ]
    TAG_GET_ATTR = None


# =====================================================================================================================




# =====================================================================================================================
# VERSION = (0, 0, 1)   # use import EXACT_OBJECTS! not *
#   from .main import *                 # INcorrect
#   from .main import EXACT_OBJECTS     # CORRECT


# =====================================================================================================================
# TEMPLATE
# from .main import (
#     # BASE
#     # AUX
#     # TYPES
#     # EXX
# )
# ---------------------------------------------------------------------------------------------------------------------
from .main import (
    # BASE
    TpInsideApi_Runner,
    TpMultyDutBase,
    # AUX
    # TYPES
    # EXX
    Exx__TcsPathNotExists,
    Exx__TcItemNotFound,
    Exx__TcItemType,
    Exx__TcSettingsIncorrect,
)
from .devices import (
    # BASE
    DeviceBase,
    DutBase,
    DevicesBreeder,
    DevicesBreeder_WithDut,
    # AUX
    DevicesBreeder_Example,
    # TYPES
    # EXX
)
from .tc import (
    # BASE
    TestCaseBase,
    # AUX
    Signals,
    # TYPES
    # EXX
)
from .tc_groups import (
    # BASE
    TcGroup_Base,
    # AUX
    # TYPES
    # EXX
)
from .tc_types import (
    # BASE
    # AUX
    # TYPES
    TYPE__RESULT_BASE,
    TYPE__RESULT_W_NORETURN,
    TYPE__RESULT_W_EXX,
    # EXX
)
from .gui import (
    # BASE
    TpGuiBase,
    # AUX
    # TYPES
    # EXX
)
from .tm import (
    # BASE
    TpTableModel,
    # AUX
    # TYPES
    # EXX
)
from .api import (
    # BASE
    TpApi_Aiohttp,
    TpApi_FastApi,
    create_app__FastApi_Tp,
    # AUX
    # TYPES
    # EXX
)

# =====================================================================================================================



from . import *

from typing import *
import time

from PyQt5.QtCore import QThread

from base_aux.funcs import *
from base_aux.classes import *
from base_aux.loggers import Logger


# =====================================================================================================================
TYPE__CMD_RESULT = Union[str, list[str]]


# =====================================================================================================================
class AnswerVariants:
    SUCCESS: str = "OK"
    FAIL: str = "FAIL"
    UNSUPPORTED: str = "UNSUPPORTED"

    ERR__NAME_CMD_OR_PARAM: str = "ERR__NAME_CMD_OR_PARAM"
    ERR__NAME_SCRIPT: str = "ERR__NAME_SCRIPT"
    ERR__VALUE_INCOMPATIBLE: str = "ERR__VALUE_INCOMPATIBLE"
    ERR__ARGS_VALIDATION: str = "ERR__ARGS_VALIDATION"

    ERR__ENCODING_OR_DEVICE: str = "ERR__ENCODING_OR_DEVICE"
    ERR__SYNTAX: str = "ERR__SYNTAX"
    ERR__PARAM_CALLING: str = "ERR__PARAM_CALLING"

    EXIT: str = "EXIT"


# =====================================================================================================================
class SerialServer_Base(Logger, QThread):
    # FIXME: PARAMS like ValueUniit/Variants is not Work/Broken!!!
    # TODO: not realized - ACCESS RULES for PARAMS - may be not need in this case of class/situation!!!
    # TODO: not realised list access - best way to use pattern "name/index" and change same access with dict "name/key"

    # SETTINGS ------------------------------------------------
    _SERIAL_CLIENT__CLS: Type[SerialClient] = SerialClient  # usually not redefines!
    SERIAL_CLIENT: SerialClient
    ADDRESS: str = None     # DON'T DEPRECATE! usually use only as exact port or keep NONE!

    HELLO_MSG__SEND_ON_START: bool = True   # don't set here on True! use it only as overwritten if needed!!!
    HELLO_MSG: list[str] = [
        "SerialServer_Base HELLO line 1",
        "SerialServer_Base hello line 2",
    ]

    PARAMS: dict[str, Union[Any, dict[Union[str, int], Any]]] = None
    ANSWER: Type[AnswerVariants] = AnswerVariants

    # AUX -----------------------------------------------------
    _STARTSWITH__CMD: str = "cmd__"
    _STARTSWITH__SCRIPT: str = "script__"

    _LIST__CMDS: list[str]
    _LIST__SCRIPTS: list[str]

    CYCLE_ACTIVE: bool = None  # active working state on ReadingWriting - used for waiting active state!

    def wait__cycle_active(self) -> None:
        """
        ALWAYS WAIT IT BEFORE START EXPLUATATION!
        """
        self.LOGGER.debug("")

        while not self.CYCLE_ACTIVE:
            time.sleep(0.5)

    @property
    def _LIST__HELP(self) -> list[str]:
        params_dump = []
        for name, value in self.PARAMS.items():
            if isinstance(value, dict):
                params_dump.append(" "*2 + f"|{name}={{")
                for name_, value_ in value.items():
                    params_dump.append(" "*4 + f"|{name_}={value_}")
            elif isinstance(value, (ValueUnit, ValueVariants)):
                params_dump.append(" "*2 + f"|{name}={repr(value)}")
            else:
                params_dump.append(" "*2 + f"|{name}={value}")

        # WORK --------------------------------
        result = [
            "="*50,
            *self.HELLO_MSG,
            "-" * 50,
            "[PARAMS]:",
            *params_dump,
            "-" * 50,
            "[CMDS]:",
            *[f"  |{name}" for name in self._LIST__CMDS],
            "-" * 50,
            "[SCRIPTS]:",
            *[f"  |{name}" for name in self._LIST__SCRIPTS],
            "-" * 50,
            "[ANSWER_VARIANTS]:",
            *[f"  |{name}" for name in dir(self.ANSWER) if not name.startswith("__")],
            "=" * 50,
        ]
        return result

    def list_param_results(self, lines: list[str]) -> list[str]:
        """
        used to show several PARAMS and CMD results (ready to pretty sent in serial port)
        cmds used as params in just case if cmd returns exact value! (not for any cmd and not used for results as Answer)
        """
        result = []
        for line in lines:
            line_parsed = LineParsed(line, _prefix_expected=self.SERIAL_CLIENT.PREFIX)
            line_result = self._cmd__(line_parsed)
            result.append(f"{line}={line_result}")

        return result

    # -----------------------------------------------------------------------------------------------------------------
    def __init__(self, params: dict[str, Any] = None):
        # FIXME: deprecate param params??? used for tests?
        super().__init__()

        if params:
            self.PARAMS = params
        elif not self.PARAMS:
            self.PARAMS = {}
        else:
            pass

        self._init__lists()

        self.SERIAL_CLIENT = self._SERIAL_CLIENT__CLS()
        self.SERIAL_CLIENT.RAISE_READ_FAIL_PATTERN = False
        self.SERIAL_CLIENT.TIMEOUT__READ = None
        if self.ADDRESS:
            self.SERIAL_CLIENT.ADDRESS = self.ADDRESS
        if not self.SERIAL_CLIENT.ADDRESS:
            self.SERIAL_CLIENT.ADDRESS = Type__AddressAutoAcceptVariant.FIRST_FREE__PAIRED   # here keep only FIRST_FREE__PAIRED! as default!

    def _init__lists(self) -> None:
        self._LIST__CMDS = []
        self._LIST__SCRIPTS = []

        for name in dir(self):
            name = name.lower()
            if name.startswith(self._STARTSWITH__CMD):
                self._LIST__CMDS.append(name.replace(self._STARTSWITH__CMD, "", 1))
            if name.startswith(self._STARTSWITH__SCRIPT):
                self._LIST__SCRIPTS.append(name.replace(self._STARTSWITH__SCRIPT, "", 1))

    # -----------------------------------------------------------------------------------------------------------------
    def run(self) -> None:
        self.LOGGER.debug("")

        if self.CYCLE_ACTIVE:
            # just for sure
            msg = f"[WARN] ALREADY STARTED={self.__class__.__name__}"
            print(msg)
            return

        if not self.SERIAL_CLIENT.connect(_raise=False):
            msg = f"[ERROR]NOT STARTED={self.__class__.__name__}"
            print(msg)
            return

        if self.HELLO_MSG__SEND_ON_START:
            additional_line = f"Started on [{self.SERIAL_CLIENT._SERIAL.port}]"
            if additional_line not in self.HELLO_MSG:
                self.HELLO_MSG.append(additional_line)
            self.SERIAL_CLIENT._write("")
            self.SERIAL_CLIENT._write("=" * 50)
            # self._execute_line("hello")
            self.SERIAL_CLIENT._write(self.HELLO_MSG)

        self._cycle__activate()

    def _cycle__activate(self) -> Never:
        self.LOGGER.debug("")

        while True:
            self.CYCLE_ACTIVE = True
            line = None
            try:
                line = self.SERIAL_CLIENT.read_line()
            except:
                self.SERIAL_CLIENT._write(self.ANSWER.ERR__ENCODING_OR_DEVICE)

            if line:
                self._execute_line(line)

    def connect(self) -> None:
        self.LOGGER.debug("")
        self.start()

    def start(self, *args, **kwargs):
        if not self.isRunning():
            super().start()
            self.wait__cycle_active()

    def disconnect(self):
        self.LOGGER.debug("")

        self.terminate()
        # self.SERIAL_CLIENT.disconnect()
        # self.CYCLE_ACTIVE = False

    def terminate(self):
        # if self.SERIAL_CLIENT.CONNECTED:
        #     self.SERIAL_CLIENT.send__(self.ANSWER.EXIT)

        self.SERIAL_CLIENT.disconnect()

        if self.isRunning():
            self.LOGGER.debug("")
            super().terminate()
        self.CYCLE_ACTIVE = False

    # -----------------------------------------------------------------------------------------------------------------
    def _execute_line(self, line: str) -> bool:
        self.LOGGER.debug("")

        line_parsed = LineParsed(line, _prefix_expected=self.SERIAL_CLIENT.PREFIX)
        cmd_result = self._cmd__(line_parsed)

        # blank line - SEND!!! because value may be BLANK!!!!
        # if not cmd_result:
        #     return True

        write_result = self.SERIAL_CLIENT._write(cmd_result)
        return write_result

    # CMD - ENTRY POINT -----------------------------------------------------------------------------------------------
    def _cmd__(self, line_parsed: LineParsed) -> TYPE__CMD_RESULT:
        self.LOGGER.debug("")

        meth_name__expected = f"{self._STARTSWITH__CMD}{line_parsed.CMD}"
        meth_name__original = IterAux().item__get_original__case_insensitive(meth_name__expected, dir(self))
        # GET METHOD --------------------
        if meth_name__original:
            meth = getattr(self, meth_name__original())
        else:
            meth = self._cmd__param_as_cmd

        # EXEC METHOD --------------------
        try:
            result = meth(line_parsed)
        except TypeError as exx:
            try:
                result = meth()
            except:
                result = self.ANSWER.FAIL
        except:
            result = self.ANSWER.FAIL

        if result is None:
            result = self.ANSWER.SUCCESS
        return result

    def _cmd__param_as_cmd(self, line_parsed: LineParsed) -> TYPE__CMD_RESULT:
        # PREPARE -------------------------------
        if line_parsed.CMD:
            line_parsed.ARGS = [line_parsed.CMD, *line_parsed.ARGS]
            line_parsed.CMD = ""

        # VALIDATE -------------------------------
        if line_parsed.ARGS and line_parsed.KWARGS:
            return self.ANSWER.ERR__ARGS_VALIDATION

        # GET -------------------------------
        if line_parsed.ARGS:
            return self.cmd__get(line_parsed)

        # SET -----------------------------------
        if line_parsed.KWARGS:
            return self.cmd__set(line_parsed)

    # CMD - PARAMS ----------------------------------------------------------------------------------------------------
    def cmd__get(self, line_parsed: LineParsed) -> TYPE__CMD_RESULT:
        """
        use only single name!!!
        """
        # ERR__ARGS_VALIDATION --------------------------------
        pass

        # PREPARE --------------------------------
        ARGS = []
        for arg in line_parsed.ARGS:
            ARGS.extend(arg.split("/"))

        # WORK --------------------------------
        param_value = IterAux().value_by_path__get(ARGS, self.PARAMS)
        if not param_value:
            return self.ANSWER.ERR__NAME_CMD_OR_PARAM

        param_value = param_value()

        # VARIANTS ------------------------------------------------------------------
        # ValueUnit -------------------------------
        if isinstance(param_value, (ValueUnit, ValueVariants)):
            return str(param_value)

        # CALLABLE -------------------------------
        # todo: add call func with remaining ARGS as func positional params??
        if callable(param_value):
            try:
                param_value = param_value()
            except:
                return self.ANSWER.ERR__PARAM_CALLING

        # NOTE - DONT return direct value! only str! cause of LIST would be assumed as list of lines! not the single line as single value!!!
        param_value = str(param_value)
        return param_value

    def cmd__set(self, line_parsed: LineParsed) -> TYPE__CMD_RESULT:
        """
        for ARGS - available only one param!
        for KWARGS - available MULTY params! fail protected!
        """
        # ERR__ARGS_VALIDATION -----------------
        if line_parsed.ARGS and line_parsed.KWARGS:
            return self.ANSWER.ERR__ARGS_VALIDATION
        if line_parsed.ARGS and line_parsed.ARGS_count() != 2:
            return self.ANSWER.ERR__ARGS_VALIDATION

        # PREPARE --------------------------------
        KWARGS = {**line_parsed.KWARGS}
        if line_parsed.ARGS:
            KWARGS = {line_parsed.ARGS[0]: line_parsed.ARGS[1]}

        # VALIDATE = check AVAILABLE TO CHANGE = exists all and not callable --------------
        for path, value_new in KWARGS.items():
            path_name__original = IterAux().path__get_original(path, self.PARAMS)
            if not path_name__original:
                return self.ANSWER.ERR__NAME_CMD_OR_PARAM

            value_old = IterAux().value_by_path__get(path, self.PARAMS)()
            if isinstance(value_old, ValueUnit):
                # NOTE: ALL CLASSES/INSTANCES ARE CALLABLE!!!
                pass
            elif isinstance(value_old, ValueVariants):
                if value_new not in value_old:
                    return self.ANSWER.ERR__VALUE_INCOMPATIBLE
            elif callable(value_old):
                return self.ANSWER.ERR__NAME_CMD_OR_PARAM

        # SET --------------
        for path, value_new in KWARGS.items():
            value_new = Strings().try_convert_to__elementary(value_new)
            value_old = IterAux().value_by_path__get(path, self.PARAMS)()
            # SET ----------
            if isinstance(value_old, (ValueUnit, ValueVariants)):
                try:
                    value_old.value = value_new
                    result = True
                except:
                    return self.ANSWER.ERR__VALUE_INCOMPATIBLE
            else:
                result = IterAux().value_by_path__set(path, value_new, self.PARAMS)

            if not result:
                return self.ANSWER.FAIL

        return self.ANSWER.SUCCESS

    # CMDS - STD ------------------------------------------------------------------------------------------------------
    def cmd__hello(self, line_parsed: LineParsed) -> TYPE__CMD_RESULT:
        # ERR__ARGS_VALIDATION --------------------------------
        pass

        # WORK --------------------------------
        return self.HELLO_MSG

    def cmd__help(self, line_parsed: LineParsed) -> TYPE__CMD_RESULT:
        # ERR__ARGS_VALIDATION --------------------------------
        pass

        # WORK --------------------------------
        return self._LIST__HELP

    def cmd__echo(self, line_parsed: LineParsed) -> TYPE__CMD_RESULT:
        # ERR__ARGS_VALIDATION --------------------------------
        pass

        # WORK --------------------------------
        return line_parsed.ORIGINAL

    # CMDS - SCRIPTS --------------------------------------------------------------------------------------------------
    def cmd__script(self, line_parsed: LineParsed) -> TYPE__CMD_RESULT:
        """
        it is as template! you can create/use your awn script-run cmd!
        """
        return self._script__(line_parsed)

    def _script__(self, line_parsed: LineParsed) -> TYPE__CMD_RESULT:
        # ERR__ARGS_VALIDATION --------------------------------
        if line_parsed.ARGS_count() == 0:
            return self.ANSWER.ERR__ARGS_VALIDATION

        # WORK --------------------------------
        meth_name__expected = f"{self._STARTSWITH__SCRIPT}{line_parsed.ARGS[0]}"
        meth_name__original = IterAux().item__get_original__case_insensitive(meth_name__expected, dir(self))
        if not meth_name__original:
            return self.ANSWER.ERR__NAME_SCRIPT

        meth = getattr(self, meth_name__original())

        # EXEC METHOD --------------------
        try:
            result = meth(line_parsed)
        except TypeError as exx:
            try:
                result = meth()
            except:
                result = self.ANSWER.FAIL
        except:
            result = self.ANSWER.FAIL

        if result is None:
            result = self.ANSWER.SUCCESS
        return result

    # def cmd__exit(self, line_parsed: LineParsed) -> TYPE__CMD_RESULT:
    #     self.disconnect()
    #     exit()


# =====================================================================================================================
class SerialServer_Echo(SerialServer_Base):
    HELLO_MSG: list[str] = [
        "SerialServer_Echo",
    ]

    def _execute_line(self, line: str) -> bool:
        write_result = self.SERIAL_CLIENT._write(line)
        return write_result


# =====================================================================================================================
class SerialServer_Example(SerialServer_Base):
    PARAMS = {
        "VAR": "",

        "STR": "str",
        "QUOTE": "str'",

        "BLANC": "",
        "ZERO": 0,

        "NONE": None,
        "TRUE": True,
        "FALSE": False,

        "INT": 1,
        "FLOAT": 1.1,

        "CALL": time.time,
        "EXX": time.strftime,

        "LIST": [0, 1, 2],
        "LIST_2": [[11]],
        "_SET": {0, 1, 2},
        "DICT_SHORT": {1: 11},
        "DICT_SHORT_2": {"HEllo": {1: 11}},
        "DICT": {
            1: 111,
            "2": 222,
            3: {
                1: 31,
                2: 32,
            },
        },
        "UNIT": ValueUnit(1, unit="V"),
        "VARIANT": ValueVariants(220, variants=[220, 380]),
    }

    def cmd__upper(self, line_parsed: LineParsed) -> TYPE__CMD_RESULT:
        # usefull for tests
        return line_parsed.ORIGINAL.upper()

    def cmd__lower(self, line_parsed: LineParsed) -> TYPE__CMD_RESULT:
        return line_parsed.ORIGINAL.lower()

    def cmd__cmd(self, line_parsed: LineParsed) -> TYPE__CMD_RESULT:
        # NOTE: NONE is equivalent for SUCCESS
        # do smth
        pass

    def cmd__cmd_no_line(self) -> TYPE__CMD_RESULT:
        # NOTE: NONE is equivalent for SUCCESS
        # do smth
        pass

    # -----------------------------------------------------------------------------------------------------------------
    def script__script1(self, line_parsed: LineParsed) -> TYPE__CMD_RESULT:
        # do smth
        pass


# =====================================================================================================================
if __name__ == "__main__":
    SerialServer_Example().run()


# =====================================================================================================================



from typing import *
from typing import Union, Any, Type, Callable, NoReturn


# =====================================================================================================================
class ValueNotExist:
    """
    DEPRECATE???
    ---------
    use direct ArgsEmpty???

    GOAL
    ----
    it is different from Default!
    there is no value!
    used when we need to change logic with not passed value!

    SPECIALLY CREATED FOR
    ---------------------
    Valid as universal validation object under cmp other objects!

    USAGE
    -----
    class Cls:
        def __init__(self, value: Any | Type[ValueNotExist] | ValueNotExist = ValueNotExist):
            self.value = value

        def __eq__(self, other):
            if self.value is ValueNotExist:
                return other is True
                # or
                return self.__class__(other).run()
            else:
                return other == self.value

        def run(self):
            return bool(self.value)

    SAME AS
    -------
    args.ArgsEmpty but single and really not defined
    """
    pass

    def __bool__(self):
        return False


# =====================================================================================================================
TYPE__VALUE_NOT_PASSED = Type[ValueNotExist] | ValueNotExist

# ---------------------------------------------------------------------------------------------------------------------
# SEE SAME BUT DIFFERS: TYPE__LAMBDA_ARGS *
TYPE__VALID_ARGS = Union[tuple, Any, None, "TYPE__ARGS_EMPTY", "TYPE__EXPLICIT"]
TYPE__VALID_KWARGS = Optional[dict[str, Any]]

TYPE__VALID_EXCEPTION = Union[Exception, Type[Exception]]
TYPE__VALID_callable_item = Callable[[...], Any | NoReturn]
TYPE__VALID_SOURCE = Union[
    Any,                                # as main idea! as already final generic
    Callable[[...], Any | NoReturn],    # as main idea! to get final generic
    # TYPE__VALID_EXCEPTION,              # fixme: hide? think no? we can pass in theory Exx! but why? for tests? why? we need it only in result! if need in tests Place it as Any!
    TYPE__VALUE_NOT_PASSED
]
TYPE__VALID_RESULT = Union[
    Any,
    TYPE__VALID_EXCEPTION,  # as main idea! instead of raise
]
# BOOL --------------------------------
TYPE__VALID_SOURCE_BOOL = Union[
    Any,                                # fixme: hide? does it need? for results like []/{}/()/0/"" think KEEP! it mean you must know that its expecting boolComparing in further logic!
    bool,                               # as main idea! as already final generic
    Callable[[...], bool | Any | NoReturn],   # as main idea! to get final generic
    # TYPE__VALID_EXCEPTION,
    TYPE__VALUE_NOT_PASSED
]
TYPE__VALID_RESULT_BOOL = Union[
    # this is when you need get only bool! raise - as False!
    bool,  # as main idea! instead of raise/exx
]
# FIXME: TODO: solve idea of BOOL!!! cant understand about Exx in here!
TYPE__VALID_RESULT_BOOL__EXX = Union[
    bool,
    TYPE__VALID_EXCEPTION,
]

# ---------------------------------------------------------------------------------------------------------------------
TYPE__VALID_VALIDATOR = Union[
    Any,    # generic final instance as expecting value - direct comparison OR comparison instance like Valid!
    # Type,   # Class as validator like Valid? fixme
    TYPE__VALID_EXCEPTION,  # direct comparison
    Callable[[Any, ...], bool | NoReturn]     # func with first param for validating source
]


# =====================================================================================================================



from typing import *
import pytest
from base_aux.stock import *
from base_aux.stock.mt import *
import MetaTrader5 as mt5


# =====================================================================================================================
class Test_mt5:
    VICTIM: Union[MT5, MT5] = type("VICTIM", (MT5,), {})

    # @classmethod
    # def setup_class(cls):
    #     pass
    #
    # @classmethod
    # def teardown_class(cls):
    #     pass
    #
    # def setup_method(self, method):
    #     # self.VICTIM = type("VICTIM", (888888888888,), {})

    # CONNECT ---------------------------------------------------------------------------------------------------------
    def test__mt5_connect(self):
        # DOUBLE CONNECT
        victim1 = self.VICTIM()
        result = mt5.last_error()
        assert result == (1, 'Success')
        assert result[0] == 1
        assert result[1] == "Success"

        victim2 = self.VICTIM()
        result = mt5.last_error()
        assert result == (1, 'Success')

        self.VICTIM.SYMBOL = "Нефть Brent"
        victim = self.VICTIM()
        assert isinstance(victim.SYMBOL, mt5.SymbolInfo)

        self.VICTIM.SYMBOL = "Нефть Brent 1234"
        try:
            victim = self.VICTIM()
        except Exception as exx:
            assert isinstance(exx, Exx__Mt5SymbolName)
        else:
            assert False

    # SORTED ---------------------------------------------------------------------------------------------------------
    def test__symbol_volume_price_get__last_day_finished(self):
        victim = self.VICTIM()
        result = victim._symbol_volume_price_get__last_day_finished(_symbol="SBER")
        print(result)

    def test__symbols_get_sorted_volume_price(self):
        victim = self.VICTIM()
        victim._symbols_get_sorted_volume_price()

    # HISTORY ---------------------------------------------------------------------------------------------------------
    def test__bars_np(self):
        victim = MT5()

        bars1 = victim.bars_get__count()
        print(bars1)            # [(1695763200, 93.89, 93.91, 93.75, 93.84, 527, 1, 2116)]
        assert bars1.shape == (1, )
        assert isinstance(bars1, np.ndarray)
        assert isinstance(bars1[0], np.void)

        print()
        print()
        print()
        print(f"{bars1[0]=}")       # (1695844800, 96.69, 96.73, 96.53, 96.56, 452, 1, 1542)
        print(f"{type(bars1[0])=}")     #<class 'numpy.void'>
        print(f"{bars1[0]['time']=}")   #1695845400
        # print(f"{bars1[0].time=}")      #AttributeError: 'numpy.void' object has no attribute 'time'
        print()
        print()
        print()

        print(bars1.dtype.fields)
        assert list(bars1.dtype.fields) == ['time', 'open', 'high', 'low', 'close', 'tick_volume', 'spread', 'real_volume']

        for elem in bars1[0]:
            print(f"{elem=}/{type(elem)}")
            """
[(1695763800, 93.83, 93.88, 93.78, 93.88, 172, 1, 723)]
elem=1695763800/<class 'numpy.int64'>
elem=93.83/<class 'numpy.float64'>
elem=93.88/<class 'numpy.float64'>
elem=93.78/<class 'numpy.float64'>
elem=93.88/<class 'numpy.float64'>
elem=172/<class 'numpy.uint64'>
elem=1/<class 'numpy.intc'>
elem=723/<class 'numpy.uint64'>
            """
            assert isinstance(elem, (np.int64, np.float64, np.uint64, np.intc, ))

        bars2 = victim.bars_get__count(2)
        print(bars2)
        assert bars2.shape == (2, )
        assert isinstance(bars2, np.ndarray)
        assert isinstance(bars2[0], np.void)

    def test__steps(self):
        victim = MT5()
        assert victim.bars_get__count(1, tf_split=1).shape == (1,)
        assert victim.bars_get__count(1, tf_split=2).shape == (1,)
        assert victim.bars_get__count(1, tf_split=3).shape == (1,)

        assert victim.bars_get__count(2, tf_split=1).shape == (2,)
        assert victim.bars_get__count(2, tf_split=2).shape == (2,)
        assert victim.bars_get__count(2, tf_split=3).shape == (2,)

    # INDICATORS ------------------------------------------------------------------------------------------------------
    def test__indicator_get__TYPE(self):
        victim = MT5()
        result = victim._indicator_get_by_obj(IndicatorParams_ADX(2, 1), return_tail=1)
        print(f"[{result=}]")
        print(f"[{type(result)=}]")
        assert isinstance(result, (int, float))

        result = victim._indicator_get_by_obj(IndicatorParams_ADX(2, 1), return_tail=2)
        print(f"[{result=}]")
        print(f"[{type(result)=}]")
        assert isinstance(result, (Type__PdSeries))

    def test__indicator_get__WMA(self):
        victim = MT5()
        results = set()
        for params in [(1, ), (10, ), (20, ), ]:
            print(f"[{params=}]")
            result = victim._indicator_get_by_obj(IndicatorParams_WMA(*params))
            print(f"[    {result=}]")
            assert isinstance(result, (int, float))
            results.add(result)
        assert len(results) > 1

        for params in [(1, ), ]:
            assert victim.indicator_WMA(params) == victim._indicator_get_by_obj(IndicatorParams_WMA(*params))

    @pytest.mark.xfail
    def test__indicator_get__STOCH(self):
        """
        it will fail because of stoch func not working correctly! always on 14/3/3!!!
        """
        victim = MT5()
        results = set()
        for params in [(14, 3, 3), (5, 1, 1), ]:
            print(f"[{params=}]")
            result = victim._indicator_get_by_obj(IndicatorParams_STOCH(*params))
            print(f"[    {result=}]")
            assert isinstance(result, (int, float))
            results.add(result)
        assert len(results) > 1

        for params in [(14, 3, 3), ]:
            assert victim.indicator_STOCH(params) == victim._indicator_get_by_obj(IndicatorParams_STOCH(*params))

    def test__indicator_get__ADX(self):
        victim = MT5()
        results = set()
        for params in [(10, 5), (5, 1), ]:
            print(f"[{params=}]")
            result = victim._indicator_get_by_obj(IndicatorParams_ADX(*params))
            print(f"[    {result=}]")
            assert isinstance(result, (int, float))
            results.add(result)
        assert len(results) > 1

        for params in [(5, 1, ), ]:
            assert victim.indicator_ADX(params) == victim._indicator_get_by_obj(IndicatorParams_ADX(*params))

    def test__indicator_get__MACD(self):
        victim = MT5()
        results = set()
        for params in [(20, 10, 20), (10, 1, 10), ]:
            print(f"[{params=}]")
            result = victim._indicator_get_by_obj(IndicatorParams_MACD(*params))
            print(f"[    {result=}]")
            assert isinstance(result, (int, float))
            results.add(result)
        assert len(results) > 1


# =====================================================================================================================
class Test_strategy:
    VICTIM: Union[MT5, MT5] = type("VICTIM", (MT5,), {})

    def test__ADX(self):
        victim = MonitorADX()
        victim.join()


# =====================================================================================================================



from base_aux.classes.singleton import *

class MySingleton(SingletonByCallMeta):
    pass

class MySingleton2(SingletonByCallMeta):
    pass

class MySingleton(metaclass=SingletonMetaCallClass):
    pass


# ===============================
# 2. access to created instances
from base_aux.classes.singleton import *

class Victim1(SingletonByCallMeta):
    attr = 1

class Victim2(SingletonByCallMeta):
    attr = 2

assert SingletonByCallMeta._SINGLETONS == []
Victim1()
assert SingletonByCallMeta._SINGLETONS == [Victim1(), ]
assert Victim1._SINGLETONS == [Victim1(), ]
assert Victim1()._SINGLETONS == [Victim1(), ]
Victim2()
assert SingletonByCallMeta._SINGLETONS == [Victim1(), Victim2(), ]


# ===============================
# 3. NOTICE: all your Singletons must be only last classes!
# don't use nesting from any Your Singletons!
from base_aux.classes.singleton import *

class MySingleton(SingletonByCallMeta):  # OK
    pass

class MySingleton2(MySingleton):  # WRONG
    pass


# =====================================================================================================================
# VERSION = (0, 0, 1)   # use import EXACT_OBJECTS! not *
#   from .main import *                 # INcorrect
#   from .main import EXACT_OBJECTS     # CORRECT
# VERSION = (0, 0, 2)   # del blank lines
# VERSION = (0, 0, 3)   # separate all types/exx into static.py!


# =====================================================================================================================
# TEMPLATE
# from .STATIC import (
#     # TYPES
#     # EXX
# )
# from .main import (
#     # BASE
#     # AUX
# )
# ---------------------------------------------------------------------------------------------------------------------
from .static import (
    Exx__AnnotNotDefined,
    Exx__NumberArithm_NoName,
    Exx__GetattrPrefix,
    Exx__GetattrPrefix_RaiseIf,
    Exx__ValueNotInVariants,
    Exx__VariantsIncompatible,
    Exx__ValueNotParsed,
    Exx__ValueUnitsIncompatible,
    Exx__IndexOverlayed,
    Exx__IndexNotSet,
    Exx__ItemNotExists,
    Exx__StartOuterNONE_UsedInStackByRecreation,
    Exx__BreederObjectList_GroupsNotGenerated,
    Exx__BreederObjectList_GroupNotExists,
    Exx__BreederObjectList_ObjCantAccessIndex,

    TYPE__LAMBDA_CONSTRUCTOR,
    TYPE__LAMBDA_ARGS,
    TYPE__LAMBDA_KWARGS,
)
from .annot_1_aux import (
    AnnotAux,
)
from .annot_3_iter_values import (
    AnnotValuesIter
)
from .annot_2_all_defined import (
    AnnotAllDefined,
)
from .annot_4_cls_keys_as_values import (
    AnnotClsKeysAsValues,
    AnnotClsKeysAsValues_Meta,
)
# ---------------------------------------------------------------------------------------------------------------------
from .cmp import (
    CmpInst,
)
from .number import (
    NumberArithmTranslateToAttr,
    TYPE__NUMBER,
)
# ---------------------------------------------------------------------------------------------------------------------
from .exceptions import ExxBool

# ---------------------------------------------------------------------------------------------------------------------
from .getattr_0_echo import (
    GetattrEcho,
    GetattrEchoSpace,
)
from .attr_1_aux import (
    AttrAux,
)
from .attr_2_anycase import (
    AttrAnycase,
)
from .getattr_3_prefix_1_inst import (
    GetattrPrefixInst,
    GetattrPrefixInst_RaiseIf,

)
from .getattr_3_prefix_2_cls import (
    GetattrPrefixCls_MetaTemplate
)

from .attr_4_lambda_call import (
    AttrLambdaCall,
)
from .lambdas import Lambda, LambdaBool, LambdaBoolReversed, LambdaTrySuccess, LambdaTryFail
# ---------------------------------------------------------------------------------------------------------------------
from .attr_3_dict_dots_1_simple import (
    DictDots,
)
from .attr_3_dict_dots_2_full import (
    DictDotsFull,
)
# ---------------------------------------------------------------------------------------------------------------------
from .middle_group import (
    ClsMiddleGroup,
)

# ---------------------------------------------------------------------------------------------------------------------
from .value_1_variants import (
    ValueVariants,
)
from .value_2_unit import (
    ValueUnit,

    UnitBase,
    UNIT_MULT__VARIANTS,
)
# ---------------------------------------------------------------------------------------------------------------------
from .valid_0_aux import ValidAux
from .valid_1_base import Valid
from .valid_2_chains import (
    ValidChains,
    TYPE__CHAINS,
)
from .valid_1_base_derivatives import (
    ValidRetry1,
    ValidRetry2,
    ValidFailStop,
    ValidFailContinue,
    ValidNoCum,
    ValidReverse,
    ValidSleep,
)
from .valid_3_regexp import (
    ValidRegExp,
)

# ---------------------------------------------------------------------------------------------------------------------
from .breeder_1_str_1_series import (
    BreederStrSeries,
)
from .breeder_1_str_2_stack import (
    BreederStrStack,
    BreederStrStack_Example,
    BreederStrStack_Example__BestUsage
)
from .breeder_2_objects import (
    BreederObjectList,
    BreederObjectList_GroupType,

    TYPE__BREED_RESULT__ITEM,
    TYPE__BREED_RESULT__GROUP,
    TYPE__BREED_RESULT__GROUPS,
)
# ---------------------------------------------------------------------------------------------------------------------
from .singleton import (
    SingletonManagerBase,
    SingletonMetaCallClass,
    SingletonByCallMeta,
    SingletonByNew,

    Exx_SingletonNestingLevels,
)


# =====================================================================================================================



from typing import *

from . import SerialClient, Type__AddressAutoAcceptVariant, SerialServer_Base, SerialServer_Example


# =====================================================================================================================
class SerialClient_FirstFree(SerialClient):
    _ADDRESS = Type__AddressAutoAcceptVariant.FIRST_FREE

    def address_forget(self) -> None:
        """
        see BaseCls!
        """
        self.ADDRESS = Type__AddressAutoAcceptVariant.FIRST_FREE

class SerialClient_FirstFree_Shorted(SerialClient):
    _ADDRESS = Type__AddressAutoAcceptVariant.FIRST_FREE__SHORTED

    def address_forget(self) -> None:
        """
        see BaseCls!
        """
        self.ADDRESS = Type__AddressAutoAcceptVariant.FIRST_FREE__SHORTED


class SerialClient_FirstFree_Paired(SerialClient):
    _ADDRESS = Type__AddressAutoAcceptVariant.FIRST_FREE__PAIRED

    def address_forget(self) -> None:
        """
        see BaseCls!
        """
        self.ADDRESS = Type__AddressAutoAcceptVariant.FIRST_FREE__PAIRED


class SerialClient_FirstFree_AnswerValid(SerialClient):
    _ADDRESS = Type__AddressAutoAcceptVariant.FIRST_FREE__ANSWER_VALID

    def address_forget(self) -> None:
        """
        see BaseCls!
        """
        self.ADDRESS = Type__AddressAutoAcceptVariant.FIRST_FREE__ANSWER_VALID


# =====================================================================================================================
class SerialClient_Emulated(SerialClient_FirstFree_Paired):
    _EMULATOR__CLS = SerialServer_Example
    _EMULATOR__START = True


# =====================================================================================================================
if __name__ == '__main__':
    victim = SerialClient_FirstFree_Shorted()
    victim.connect()


# =====================================================================================================================
